from django.urls import reverse_lazy
from rest_framework.test import APITestCase

from tests.testapp.models import (
    Book,
    Instructor,
    Course,
    Phone,
    Student,
    Post,
    Attachment,
)
from tests.testapp.serializers import (
    WritableCourseSerializer,
    WritableStudentSerializer,
)


class DataQueryingTests(APITestCase):
    def setUp(self):
        self.book1 = Book.objects.create(
            title="Advanced Data Structures", author="S.Mobit"
        )
        self.book2 = Book.objects.create(
            title="Basic Data Structures", author="S.Mobit"
        )

        self.course = Course.objects.create(name="Data Structures", code="CS210")

        self.course.books.set([self.book1, self.book2])

        self.student = Student.objects.create(name="Yezy", age=24, course=self.course)

        self.phone1 = Phone.objects.create(
            number="076711110", type="Office", student=self.student
        )
        self.phone2 = Phone.objects.create(
            number="073008880", type="Home", student=self.student
        )

    def add_second_student(self):
        """
        Adds an additional student with their own course, course books, and phone numbers.
        """
        book1 = Book.objects.create(title="Algorithm Design", author="S.Mobit")
        book2 = Book.objects.create(title="Proving Algorithms", author="S.Mobit")

        course = Course.objects.create(name="Algorithms", code="CS260")

        course.books.set([book1, book2])

        student = Student.objects.create(name="Tyler", age=25, course=course)

        Phone.objects.create(number="075711110", type="Office", student=student)
        Phone.objects.create(number="073008880", type="Home", student=student)

        student.refresh_from_db()

        return student

    def tearDown(self):
        Book.objects.all().delete()
        Course.objects.all().delete()
        Student.objects.all().delete()

    # *************** requestless tests **************

    def test_querying_data_without_request(self):
        serializer = WritableCourseSerializer(
            self.course, query="{name, code, books{title}}"
        )

        self.assertEqual(
            serializer.data,
            {
                "name": "Data Structures",
                "code": "CS210",
                "books": [
                    {"title": "Advanced Data Structures"},
                    {"title": "Basic Data Structures"},
                ],
            },
        )

    # *************** retrieve tests **************

    def test_retrieve_with_flat_query(self):
        url = reverse_lazy("book-detail", args=[self.book1.id])
        response = self.client.get(url + "?query={title, author,}", format="json")

        self.assertEqual(
            response.data,
            {"title": "Advanced Data Structures", "author": "S.Mobit"},
        )

    def test_retrieve_with_nested_flat_query(self):
        url = reverse_lazy("student-detail", args=[self.student.id])
        response = self.client.get(
            url + "?query={name, age, course{name}}", format="json"
        )

        self.assertEqual(
            response.data,
            {"name": "Yezy", "age": 24, "course": {"name": "Data Structures"}},
        )

    def test_retrieve_with_nested_iterable_query(self):
        url = reverse_lazy("course-detail", args=[self.course.id])
        response = self.client.get(
            url + "?query={name, code, books{title}}", format="json"
        )

        self.assertEqual(
            response.data,
            {
                "name": "Data Structures",
                "code": "CS210",
                "books": [
                    {"title": "Advanced Data Structures"},
                    {"title": "Basic Data Structures"},
                ],
            },
        )

    def test_retrieve_with_nested_flat_query_and_aliased_field(self):
        url = reverse_lazy("student-detail", args=[self.student.id])
        response = self.client.get(
            url + "?query={name, age, programme: course{name}}", format="json"
        )

        self.assertEqual(
            response.data,
            {"name": "Yezy", "age": 24, "programme": {"name": "Data Structures"}},
        )

    def test_retrieve_with_nested_iterable_query_and_aliased_field(self):
        url = reverse_lazy("course-detail", args=[self.course.id])
        response = self.client.get(
            url + "?query={name, code, readings: books{bookTitle: title}}",
            format="json",
        )

        self.assertEqual(
            response.data,
            {
                "name": "Data Structures",
                "code": "CS210",
                "readings": [
                    {"bookTitle": "Advanced Data Structures"},
                    {"bookTitle": "Basic Data Structures"},
                ],
            },
        )

    def test_retrieve_with_disable_dynamic_fields_enabled(self):
        url = reverse_lazy(
            "course_with_disable_dynamic_fields_kwarg-detail", args=[self.course.id]
        )
        response = self.client.get(
            url + "?query={name, code, books{title}}", format="json"
        )

        self.assertEqual(
            response.data,
            {
                "name": "Data Structures",
                "code": "CS210",
                "books": [
                    {"title": "Advanced Data Structures", "author": "S.Mobit"},
                    {"title": "Basic Data Structures", "author": "S.Mobit"},
                ],
            },
        )

    def test_retrieve_reverse_relation_with_nested_iterable_query(self):
        url = reverse_lazy("student-detail", args=[self.student.id])
        response = self.client.get(
            url + "?query={name, age, phone_numbers{number}}", format="json"
        )

        self.assertEqual(
            response.data,
            {
                "name": "Yezy",
                "age": 24,
                "phone_numbers": [{"number": "076711110"}, {"number": "073008880"}],
            },
        )

    def test_retrieve_with_exclude_operator_applied_at_the_top_level(self):
        url = reverse_lazy("course-detail", args=[self.course.id])
        response = self.client.get(url + "?query={-books}", format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Data Structures",
                "code": "CS210",
            },
        )

    def test_retrieve_with_exclude_operator_applied_on_a_nested_field(self):
        url = reverse_lazy("student-detail", args=[self.student.id])
        response = self.client.get(
            url + "?query={name age phone_numbers{-number -student}}", format="json"
        )

        self.assertEqual(
            response.data,
            {
                "name": "Yezy",
                "age": 24,
                "phone_numbers": [{"type": "Office"}, {"type": "Home"}],
            },
        )

    def test_retrieve_with_exclude_operator_applied_with_include_all_operator(self):
        url = reverse_lazy("student-detail", args=[self.student.id])
        response = self.client.get(
            url + "?query={*, -course, phone: phone_numbers{-number, -student}}",
            format="json",
        )

        self.assertEqual(
            response.data,
            {
                "name": "Yezy",
                "age": 24,
                "phone": [{"type": "Office"}, {"type": "Home"}],
            },
        )

    def test_retrieve_with_exclude_operator_applied_on_including_and_excluding_query(
        self,
    ):
        url = reverse_lazy("student-detail", args=[self.student.id])
        response = self.client.get(
            url + "?query={-course, studentAge : age, phone: phone_numbers{type}}",
            format="json",
        )

        self.assertEqual(
            response.data,
            {
                "name": "Yezy",
                "studentAge": 24,
                "phone": [{"type": "Office"}, {"type": "Home"}],
            },
        )

    def test_retrieve_with_exclude_operator_applied_at_the_top_level_and_expand_a_nested_field(
        self,
    ):
        url = reverse_lazy("student-detail", args=[self.student.id])
        response = self.client.get(
            url + "?query={-age, -course, phone_numbers{number}}", format="json"
        )

        self.assertEqual(
            response.data,
            {
                "name": "Yezy",
                "phone_numbers": [{"number": "076711110"}, {"number": "073008880"}],
            },
        )

    def test_retrieve_with_include_all_operator_applied_at_the_top_level_and_expand_a_nested_field(
        self,
    ):
        url = reverse_lazy("student-detail", args=[self.student.id])
        response = self.client.get(
            url + "?query={*, phone_numbers{-type, -student}}", format="json"
        )

        self.assertEqual(
            response.data,
            {
                "name": "Yezy",
                "age": 24,
                "course": {
                    "name": "Data Structures",
                    "code": "CS210",
                    "books": [
                        {"title": "Advanced Data Structures", "author": "S.Mobit"},
                        {"title": "Basic Data Structures", "author": "S.Mobit"},
                    ],
                },
                "phone_numbers": [{"number": "076711110"}, {"number": "073008880"}],
            },
        )

    def test_retrieve_with_exclude_operator_and_include_all_operator_applied_at_the_top_level(
        self,
    ):
        url = reverse_lazy("student-detail", args=[self.student.id])
        response = self.client.get(
            url + "?query={*, -age, phone_numbers{-type, -student}}", format="json"
        )

        self.assertEqual(
            response.data,
            {
                "name": "Yezy",
                "course": {
                    "name": "Data Structures",
                    "code": "CS210",
                    "books": [
                        {"title": "Advanced Data Structures", "author": "S.Mobit"},
                        {"title": "Basic Data Structures", "author": "S.Mobit"},
                    ],
                },
                "phone_numbers": [{"number": "076711110"}, {"number": "073008880"}],
            },
        )

    def test_retrieve_with_exclude_and_include_all_operators_being_a_little_forgiving_on_the_syntax(
        self,
    ):
        url = reverse_lazy("student-detail", args=[self.student.id])
        # We could remove `age` field and * under `phone_numbers` field and get the same results
        response = self.client.get(
            url + "?query={*, age, phone_numbers{*,-type, -student}}", format="json"
        )

        self.assertEqual(
            response.data,
            {
                "name": "Yezy",
                "age": 24,
                "course": {
                    "name": "Data Structures",
                    "code": "CS210",
                    "books": [
                        {"title": "Advanced Data Structures", "author": "S.Mobit"},
                        {"title": "Basic Data Structures", "author": "S.Mobit"},
                    ],
                },
                "phone_numbers": [{"number": "076711110"}, {"number": "073008880"}],
            },
        )

    def test_retrieve_without_query_param(self):
        url = reverse_lazy("student-detail", args=[self.student.id])
        response = self.client.get(url, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Yezy",
                "age": 24,
                "course": {
                    "name": "Data Structures",
                    "code": "CS210",
                    "books": [
                        {"title": "Advanced Data Structures", "author": "S.Mobit"},
                        {"title": "Basic Data Structures", "author": "S.Mobit"},
                    ],
                },
                "phone_numbers": [
                    {"number": "076711110", "type": "Office", "student": 1},
                    {"number": "073008880", "type": "Home", "student": 1},
                ],
            },
        )

    def test_retrieve_eager_loading_view_mixin(self):
        """
        Ensure that we apply our prefetching or joins when we explicitly ask for fields in the
        mapping.
        """
        non_mixin_url = reverse_lazy("student-detail", args=[self.student.id])
        mixin_url = reverse_lazy("student_eager_loading-detail", args=[self.student.id])

        # Will need to fetch the course and books on serialization.
        with self.assertNumQueries(3):
            response = self.client.get(
                non_mixin_url + "?query={name, age, course{name, books}}", format="json"
            )
            self.assertEqual(
                response.data,
                {
                    "name": "Yezy",
                    "age": 24,
                    "course": {
                        "name": "Data Structures",
                        "books": [
                            {"title": "Advanced Data Structures", "author": "S.Mobit"},
                            {"title": "Basic Data Structures", "author": "S.Mobit"},
                        ],
                    },
                },
            )

        # Should select_related the course and prefetch the books.
        with self.assertNumQueries(2):
            response = self.client.get(
                mixin_url + "?query={name, age, program{name, books}}", format="json"
            )
            self.assertEqual(
                response.data,
                {
                    "name": "Yezy",
                    "age": 24,
                    "program": {
                        "name": "Data Structures",
                        "books": [
                            {"title": "Advanced Data Structures", "author": "S.Mobit"},
                            {"title": "Basic Data Structures", "author": "S.Mobit"},
                        ],
                    },
                },
            )

    def test_retrieve_eager_loading_view_mixin_ignored(self):
        """
        Ensure that we do not apply our joining/prefetching if the mapped fields aren't present.
        """
        url = reverse_lazy("student_eager_loading-detail", args=[self.student.id])

        # Make sure that no additional prefetching or select_related are run if we don't ask for
        # nested values.
        with self.assertNumQueries(1):
            response = self.client.get(url + "?query={name, age}", format="json")
            self.assertEqual(
                response.data,
                {
                    "name": "Yezy",
                    "age": 24,
                },
            )

    def test_retrieve_eager_loading_view_mixin_implicit(self):
        """
        Test that we implicitly apply our nested prefetching, since the field is present.
        """
        url = reverse_lazy("student_eager_loading-detail", args=[self.student.id])

        # This would be 3 without doing a select_related.
        with self.assertNumQueries(2):
            response = self.client.get(
                url + "?query={name, age, program}", format="json"
            )
            self.assertEqual(
                response.data,
                {
                    "name": "Yezy",
                    "age": 24,
                    "program": {
                        "name": "Data Structures",
                        "code": "CS210",
                        "books": [
                            {"title": "Advanced Data Structures", "author": "S.Mobit"},
                            {"title": "Basic Data Structures", "author": "S.Mobit"},
                        ],
                    },
                },
            )

    def test_retrieve_eager_loading_view_mixin_all_exclude(self):
        """
        Test that we implicitly apply our prefetching with * and exclusion.
        """
        url = reverse_lazy("student_eager_loading-detail", args=[self.student.id])

        # This would be 3 without doing a select_related.
        with self.assertNumQueries(2):
            response = self.client.get(
                url + "?query={*, -phone_numbers, program{*, books{title}}}",
                format="json",
            )
            self.assertEqual(
                response.data,
                {
                    "name": "Yezy",
                    "age": 24,
                    "program": {
                        "name": "Data Structures",
                        "code": "CS210",
                        "books": [
                            {"title": "Advanced Data Structures"},
                            {"title": "Basic Data Structures"},
                        ],
                    },
                },
            )

    # *************** list tests **************

    def test_list_with_flat_query(self):
        url = reverse_lazy("book-list")
        response = self.client.get(url + "?query={title, author}", format="json")

        self.assertEqual(
            response.data,
            [
                {"title": "Advanced Data Structures", "author": "S.Mobit"},
                {"title": "Basic Data Structures", "author": "S.Mobit"},
            ],
        )

    def test_list_with_nested_flat_query(self):
        url = reverse_lazy("student-list")
        response = self.client.get(
            url + "?query={name, age, course{name}}", format="json"
        )

        self.assertEqual(
            response.data,
            [{"name": "Yezy", "age": 24, "course": {"name": "Data Structures"}}],
        )

    def test_list_with_nested_iterable_query(self):
        url = reverse_lazy("course-list")
        response = self.client.get(
            url + "?query={name, code, books{title}}", format="json"
        )

        self.assertEqual(
            response.data,
            [
                {
                    "name": "Data Structures",
                    "code": "CS210",
                    "books": [
                        {"title": "Advanced Data Structures"},
                        {"title": "Basic Data Structures"},
                    ],
                }
            ],
        )

    def test_list_reverse_relation_with_nested_iterable_query(self):
        url = reverse_lazy("student-list")
        response = self.client.get(
            url + "?query={name, age, phone_numbers{number}}", format="json"
        )

        self.assertEqual(
            response.data,
            [
                {
                    "name": "Yezy",
                    "age": 24,
                    "phone_numbers": [{"number": "076711110"}, {"number": "073008880"}],
                }
            ],
        )

    def test_list_with_nested_flat_and_deep_iterable_query(self):
        url = reverse_lazy("student-list")
        response = self.client.get(
            url + "?query={name, age, course{name, books{title}}}", format="json"
        )

        self.assertEqual(
            response.data,
            [
                {
                    "name": "Yezy",
                    "age": 24,
                    "course": {
                        "name": "Data Structures",
                        "books": [
                            {"title": "Advanced Data Structures"},
                            {"title": "Basic Data Structures"},
                        ],
                    },
                }
            ],
        )

    def test_list_without_query_param(self):
        url = reverse_lazy("student-list")
        response = self.client.get(url, format="json")

        self.assertEqual(
            response.data,
            [
                {
                    "name": "Yezy",
                    "age": 24,
                    "course": {
                        "name": "Data Structures",
                        "code": "CS210",
                        "books": [
                            {"title": "Advanced Data Structures", "author": "S.Mobit"},
                            {"title": "Basic Data Structures", "author": "S.Mobit"},
                        ],
                    },
                    "phone_numbers": [
                        {"number": "076711110", "type": "Office", "student": 1},
                        {"number": "073008880", "type": "Home", "student": 1},
                    ],
                }
            ],
        )

    def test_list_with_nested_field_without_expanding(self):
        url = reverse_lazy("student-list")
        response = self.client.get(
            url + "?query={name, age, course{name, books}}", format="json"
        )

        self.assertEqual(
            response.data,
            [
                {
                    "name": "Yezy",
                    "age": 24,
                    "course": {
                        "name": "Data Structures",
                        "books": [
                            {"title": "Advanced Data Structures", "author": "S.Mobit"},
                            {"title": "Basic Data Structures", "author": "S.Mobit"},
                        ],
                    },
                }
            ],
        )

    def test_list_with_serializer_field_kwarg(self):
        url = reverse_lazy("course_with_field_kwarg-list")
        response = self.client.get(url, format="json")

        self.assertEqual(
            response.data,
            [
                {
                    "name": "Data Structures",
                    "code": "CS210",
                    "books": [
                        {"title": "Advanced Data Structures"},
                        {"title": "Basic Data Structures"},
                    ],
                }
            ],
        )

    def test_list_with_serializer_exclude_kwarg(self):
        url = reverse_lazy("course_with_exclude_kwarg-list")
        response = self.client.get(url, format="json")

        self.assertEqual(
            response.data,
            [
                {
                    "name": "Data Structures",
                    "code": "CS210",
                    "books": [
                        {"title": "Advanced Data Structures"},
                        {"title": "Basic Data Structures"},
                    ],
                }
            ],
        )

    def test_list_with_serializer_return_pk_kwarg(self):
        url = reverse_lazy("course_with_returnpk_kwarg-list")
        response = self.client.get(url, format="json")

        self.assertEqual(
            response.data,
            [{"name": "Data Structures", "code": "CS210", "books": [1, 2]}],
        )

    def test_list_with_aliased_books(self):
        url = reverse_lazy("course_with_aliased_books-list")
        response = self.client.get(url + "?query={name, tomes{title}}", format="json")

        self.assertEqual(
            response.data,
            [
                {
                    "name": "Data Structures",
                    "tomes": [
                        {"title": "Advanced Data Structures"},
                        {"title": "Basic Data Structures"},
                    ],
                }
            ],
        )

    def test_list_with_dynamic_serializer_method_field(self):
        url = reverse_lazy("course_with_dynamic_serializer_method_field-list")
        response = self.client.get(url + "?query={name, tomes}", format="json")

        self.assertEqual(
            response.data,
            [
                {
                    "name": "Data Structures",
                    "tomes": [
                        {"title": "Advanced Data Structures", "author": "S.Mobit"},
                        {"title": "Basic Data Structures", "author": "S.Mobit"},
                    ],
                }
            ],
        )

    def test_list_with_expanded_dynamic_serializer_method_field(self):
        url = reverse_lazy("course_with_dynamic_serializer_method_field-list")
        response = self.client.get(url + "?query={name, tomes{title}}", format="json")

        self.assertEqual(
            response.data,
            [
                {
                    "name": "Data Structures",
                    "tomes": [
                        {"title": "Advanced Data Structures"},
                        {"title": "Basic Data Structures"},
                    ],
                }
            ],
        )

    def test_list_with_dynamic_serializer_method_field_and_aliased_fields(self):
        url = reverse_lazy("course_with_dynamic_serializer_method_field-list")
        response = self.client.get(
            url + "?query={courseName: name, tomes{*, bookTitle: title}}", format="json"
        )

        self.assertEqual(
            response.data,
            [
                {
                    "courseName": "Data Structures",
                    "tomes": [
                        {"bookTitle": "Advanced Data Structures", "author": "S.Mobit"},
                        {"bookTitle": "Basic Data Structures", "author": "S.Mobit"},
                    ],
                }
            ],
        )

    def test_list_with_expanded_dynamic_serializer_method_field_and_aliased_fields(
        self,
    ):
        url = reverse_lazy("course_with_dynamic_serializer_method_field-list")
        response = self.client.get(
            url + "?query={name, tomes{bookTitle: title}}", format="json"
        )

        self.assertEqual(
            response.data,
            [
                {
                    "name": "Data Structures",
                    "tomes": [
                        {"bookTitle": "Advanced Data Structures"},
                        {"bookTitle": "Basic Data Structures"},
                    ],
                }
            ],
        )

    def test_list_with_dynamic_serializer_method_field_without_query_this_tests_query_and_parsed_query_kwargs_too(
        self,
    ):
        """
        Test that the DynamicSerializerMethodField works without a query present.
        """
        url = reverse_lazy("course_with_dynamic_serializer_method_field-list")
        response = self.client.get(url, format="json")

        self.assertEqual(
            response.data,
            [
                {
                    "name": "Data Structures",
                    "code": "CS210",
                    "tomes": [
                        {"title": "Advanced Data Structures", "author": "S.Mobit"},
                        {"title": "Basic Data Structures", "author": "S.Mobit"},
                    ],
                    "related_books": [
                        {"title": "Advanced Data Structures"},
                        {"title": "Basic Data Structures"},
                    ],
                }
            ],
        )

    def test_list_eager_loading_view_mixin(self):
        """
        Ensure that we apply our prefetching or joins when we explicitly ask for fields in the
        mapping.
        """
        non_mixin_url = reverse_lazy("student-list")
        mixin_url = reverse_lazy("student_eager_loading-list")

        self.add_second_student()

        # This fetches both course names and books for each, so 5 queries total.
        with self.assertNumQueries(5):
            response = self.client.get(
                non_mixin_url + "?query={name, age, course{name, books}}", format="json"
            )
            self.assertEqual(
                response.data,
                [
                    {
                        "name": "Yezy",
                        "age": 24,
                        "course": {
                            "name": "Data Structures",
                            "books": [
                                {
                                    "title": "Advanced Data Structures",
                                    "author": "S.Mobit",
                                },
                                {"title": "Basic Data Structures", "author": "S.Mobit"},
                            ],
                        },
                    },
                    {
                        "name": "Tyler",
                        "age": 25,
                        "course": {
                            "name": "Algorithms",
                            "books": [
                                {"title": "Algorithm Design", "author": "S.Mobit"},
                                {"title": "Proving Algorithms", "author": "S.Mobit"},
                            ],
                        },
                    },
                ],
            )

        # This fetches the course information with the student and prefetches once for the books.
        with self.assertNumQueries(2):
            response = self.client.get(
                mixin_url + "?query={name, age, program{name, books}}", format="json"
            )
            self.assertEqual(
                response.data,
                [
                    {
                        "name": "Yezy",
                        "age": 24,
                        "program": {
                            "name": "Data Structures",
                            "books": [
                                {
                                    "title": "Advanced Data Structures",
                                    "author": "S.Mobit",
                                },
                                {"title": "Basic Data Structures", "author": "S.Mobit"},
                            ],
                        },
                    },
                    {
                        "name": "Tyler",
                        "age": 25,
                        "program": {
                            "name": "Algorithms",
                            "books": [
                                {"title": "Algorithm Design", "author": "S.Mobit"},
                                {"title": "Proving Algorithms", "author": "S.Mobit"},
                            ],
                        },
                    },
                ],
            )

    def test_list_eager_loading_view_mixin_with_aliased_fields(self):
        """
        Ensure that we apply our prefetching or joins when we explicitly ask for fields in the
        mapping.
        """
        url = reverse_lazy("student_eager_loading-list")

        self.add_second_student()

        # This fetches the course information with the student and prefetches once for the books.
        with self.assertNumQueries(2):
            response = self.client.get(
                url + "?query={name, age, prog: program{name, readings: books}}",
                format="json",
            )
            self.assertEqual(
                response.data,
                [
                    {
                        "name": "Yezy",
                        "age": 24,
                        "prog": {
                            "name": "Data Structures",
                            "readings": [
                                {
                                    "title": "Advanced Data Structures",
                                    "author": "S.Mobit",
                                },
                                {"title": "Basic Data Structures", "author": "S.Mobit"},
                            ],
                        },
                    },
                    {
                        "name": "Tyler",
                        "age": 25,
                        "prog": {
                            "name": "Algorithms",
                            "readings": [
                                {"title": "Algorithm Design", "author": "S.Mobit"},
                                {"title": "Proving Algorithms", "author": "S.Mobit"},
                            ],
                        },
                    },
                ],
            )

    def test_list_eager_loading_view_mixin_ignored(self):
        """
        Ensure that we do not apply our joining/prefetching if the mapped fields aren't present.
        """
        url = reverse_lazy("student_eager_loading-list")
        self.add_second_student()

        # Make sure that no additional prefetching or select_related are run if we don't ask for
        # nested values.
        with self.assertNumQueries(1):
            response = self.client.get(url + "?query={name, age}", format="json")
            self.assertEqual(
                response.data,
                [
                    {
                        "name": "Yezy",
                        "age": 24,
                    },
                    {
                        "name": "Tyler",
                        "age": 25,
                    },
                ],
            )

    def test_list_eager_loading_view_mixin_implicit(self):
        """
        Test that we implicitly apply our nested prefetching, since the field is present.
        """
        url = reverse_lazy("student_eager_loading-list")
        self.add_second_student()

        # This would be 5 without doing a select_related and prefetch_related.
        with self.assertNumQueries(2):
            response = self.client.get(
                url + "?query={name, age, program}", format="json"
            )
            self.assertEqual(
                response.data,
                [
                    {
                        "name": "Yezy",
                        "age": 24,
                        "program": {
                            "name": "Data Structures",
                            "code": "CS210",
                            "books": [
                                {
                                    "title": "Advanced Data Structures",
                                    "author": "S.Mobit",
                                },
                                {"title": "Basic Data Structures", "author": "S.Mobit"},
                            ],
                        },
                    },
                    {
                        "name": "Tyler",
                        "age": 25,
                        "program": {
                            "name": "Algorithms",
                            "code": "CS260",
                            "books": [
                                {"title": "Algorithm Design", "author": "S.Mobit"},
                                {"title": "Proving Algorithms", "author": "S.Mobit"},
                            ],
                        },
                    },
                ],
            )

    def test_list_eager_loading_view_mixin_all_exclude(self):
        """
        Test that we implicitly apply our prefetching with * and exclusion.
        """
        url = reverse_lazy("student_eager_loading-list")
        self.add_second_student()

        # This would be 5 without doing a select_related and prefetch_related.
        with self.assertNumQueries(2):
            response = self.client.get(
                url + "?query={*, -phone_numbers, program{*, books{title}}}",
                format="json",
            )
            self.assertEqual(
                response.data,
                [
                    {
                        "name": "Yezy",
                        "age": 24,
                        "program": {
                            "name": "Data Structures",
                            "code": "CS210",
                            "books": [
                                {"title": "Advanced Data Structures"},
                                {"title": "Basic Data Structures"},
                            ],
                        },
                    },
                    {
                        "name": "Tyler",
                        "age": 25,
                        "program": {
                            "name": "Algorithms",
                            "code": "CS260",
                            "books": [
                                {"title": "Algorithm Design"},
                                {"title": "Proving Algorithms"},
                            ],
                        },
                    },
                ],
            )

    def test_list_eager_loading_mixin_without_query_param(self):
        url = reverse_lazy("student_eager_loading-list")
        self.add_second_student()

        with self.assertNumQueries(3):
            response = self.client.get(url, format="json")
            self.assertEqual(
                response.data,
                [
                    {
                        "name": "Yezy",
                        "age": 24,
                        "program": {
                            "name": "Data Structures",
                            "code": "CS210",
                            "books": [
                                {
                                    "title": "Advanced Data Structures",
                                    "author": "S.Mobit",
                                },
                                {"title": "Basic Data Structures", "author": "S.Mobit"},
                            ],
                        },
                        "phone_numbers": [
                            {"number": "076711110", "type": "Office", "student": 1},
                            {"number": "073008880", "type": "Home", "student": 1},
                        ],
                    },
                    {
                        "name": "Tyler",
                        "age": 25,
                        "program": {
                            "name": "Algorithms",
                            "code": "CS260",
                            "books": [
                                {"title": "Algorithm Design", "author": "S.Mobit"},
                                {"title": "Proving Algorithms", "author": "S.Mobit"},
                            ],
                        },
                        "phone_numbers": [
                            {"number": "075711110", "type": "Office", "student": 2},
                            {"number": "073008880", "type": "Home", "student": 2},
                        ],
                    },
                ],
            )

    def test_list_eager_loading_mixin_with_exclude_operator_but_without_wildcard(self):
        url = reverse_lazy("student_eager_loading-list")
        self.add_second_student()

        with self.assertNumQueries(2):
            response = self.client.get(url + "?query={-phone_numbers}", format="json")
            self.assertEqual(
                response.data,
                [
                    {
                        "name": "Yezy",
                        "age": 24,
                        "program": {
                            "name": "Data Structures",
                            "code": "CS210",
                            "books": [
                                {
                                    "title": "Advanced Data Structures",
                                    "author": "S.Mobit",
                                },
                                {"title": "Basic Data Structures", "author": "S.Mobit"},
                            ],
                        },
                    },
                    {
                        "name": "Tyler",
                        "age": 25,
                        "program": {
                            "name": "Algorithms",
                            "code": "CS260",
                            "books": [
                                {"title": "Algorithm Design", "author": "S.Mobit"},
                                {"title": "Proving Algorithms", "author": "S.Mobit"},
                            ],
                        },
                    },
                ],
            )

    def test_list_eager_loading_mixin_with_empty_query_param(self):
        url = reverse_lazy("student_eager_loading-list")
        self.add_second_student()

        with self.assertNumQueries(1):
            response = self.client.get(url + "?query={}", format="json")

            self.assertEqual(
                response.data,
                [
                    {},
                    {},
                ],
            )

    def test_list_eager_loading_mixin_with_prefetch_object_outside_of_list(self):
        """
        Test that a Prefetch object can be provided in the mapping outside of a list.
        """
        url = reverse_lazy("student_eager_loading_prefetch-list")
        self.add_second_student()

        with self.assertNumQueries(2):
            response = self.client.get(url + "?query={program}", format="json")
            self.assertEqual(
                response.data,
                [
                    {
                        "program": {
                            "name": "Data Structures",
                            "code": "CS210",
                            "books": [
                                {
                                    "title": "Advanced Data Structures",
                                    "author": "S.Mobit",
                                },
                                {"title": "Basic Data Structures", "author": "S.Mobit"},
                            ],
                        }
                    },
                    {
                        "program": {
                            "name": "Algorithms",
                            "code": "CS260",
                            "books": [
                                {"title": "Algorithm Design", "author": "S.Mobit"},
                                {"title": "Proving Algorithms", "author": "S.Mobit"},
                            ],
                        }
                    },
                ],
            )

    def test_list_eager_loading_mixin_with_prefetch_object_in_list(self):
        """
        Test that a Prefetch object can be provided in the mapping inside of a list.s
        """
        url = reverse_lazy("student_eager_loading_prefetch-list")
        self.add_second_student()

        with self.assertNumQueries(2):
            response = self.client.get(url + "?query={phone_numbers}", format="json")
            self.assertEqual(
                response.data,
                [
                    {
                        "phone_numbers": [
                            {"number": "076711110", "type": "Office", "student": 1},
                            {"number": "073008880", "type": "Home", "student": 1},
                        ]
                    },
                    {
                        "phone_numbers": [
                            {"number": "075711110", "type": "Office", "student": 2},
                            {"number": "073008880", "type": "Home", "student": 2},
                        ]
                    },
                ],
            )

    def test_list_with_auto_apply_eager_loading_set_false(self):
        """
        Test that a Prefetch object can be provided in the mapping inside of a list.s
        """
        url = reverse_lazy("student_auto_apply_eager_loading-list")
        self.add_second_student()

        with self.assertNumQueries(7):
            response = self.client.get(url, format="json")
            self.assertEqual(
                response.data,
                [
                    {
                        "name": "Yezy",
                        "age": 24,
                        "program": {
                            "name": "Data Structures",
                            "code": "CS210",
                            "books": [
                                {
                                    "title": "Advanced Data Structures",
                                    "author": "S.Mobit",
                                },
                                {"title": "Basic Data Structures", "author": "S.Mobit"},
                            ],
                        },
                        "phone_numbers": [
                            {"number": "076711110", "type": "Office", "student": 1},
                            {"number": "073008880", "type": "Home", "student": 1},
                        ],
                    },
                    {
                        "name": "Tyler",
                        "age": 25,
                        "program": {
                            "name": "Algorithms",
                            "code": "CS260",
                            "books": [
                                {"title": "Algorithm Design", "author": "S.Mobit"},
                                {"title": "Proving Algorithms", "author": "S.Mobit"},
                            ],
                        },
                        "phone_numbers": [
                            {"number": "075711110", "type": "Office", "student": 2},
                            {"number": "073008880", "type": "Home", "student": 2},
                        ],
                    },
                ],
            )

    def test_list_on_arguments_with_no_quoted_values(self):
        url = reverse_lazy("student-list")
        response = self.client.get(
            url + '?query=(name: "Yezy", age: 20){name, age, course{name}}',
            format="json",
        )

        self.assertEqual(response.data, [])

    def test_list_on_arguments_with_single_quoted_string_as_value(self):
        url = reverse_lazy("student-list")
        response = self.client.get(
            url + "?query=(name: 'Yezy', age: '20',){name, age, course{name,},}",
            format="json",
        )

        self.assertEqual(response.data, [])

    def test_list_on_arguments_with_double_quoted_string_as_value(self):
        url = reverse_lazy("student-list")
        response = self.client.get(
            url + '?query=(name: "Yezy", age: 20){name, age, course{name}}',
            format="json",
        )

        self.assertEqual(response.data, [])

    def test_list_with_nested_arguments(self):
        url = reverse_lazy("student-list")
        response = self.client.get(
            url + '?query={name, age, course(code: "CS50"){name}}', format="json"
        )

        self.assertEqual(response.data, [])

    def test_list_with_applied_filter_on_nested_aliased_field(self):
        url = reverse_lazy("student-list")
        response = self.client.get(
            url + '?query={name, age, programme: course(code: "CS015"){code}}',
            format="json",
        )

        self.assertEqual(response.data, [])

    def test_list_with_applied_filter_on_very_deeply_nested_aliased_field(self):
        url = reverse_lazy("student-list")
        response = self.client.get(
            url
            + '?query={name, age, program: course{readings: books(author: "Y.Mobit"){author}}}',
            format="json",
        )

        self.assertEqual(response.data, [])

    def test_options_on_serializer_with_fields_kwarg(self):
        url = reverse_lazy("course_with_field_kwarg-list")
        response = self.client.options(url, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Course With Fields Kwarg List",
                "description": "",
                "renders": ["application/json", "text/html"],
                "parses": [
                    "application/json",
                    "application/x-www-form-urlencoded",
                    "multipart/form-data",
                ],
                "actions": {
                    "POST": {
                        "name": {
                            "type": "string",
                            "required": True,
                            "read_only": False,
                            "label": "Name",
                            "max_length": 50,
                        },
                        "code": {
                            "type": "string",
                            "required": True,
                            "read_only": False,
                            "label": "Code",
                            "max_length": 30,
                        },
                        "books": {
                            "type": "field",
                            "required": False,
                            "read_only": True,
                            "label": "Books",
                            "child": {
                                "type": "nested object",
                                "required": False,
                                "read_only": True,
                                "children": {
                                    "title": {
                                        "type": "string",
                                        "required": True,
                                        "read_only": False,
                                        "label": "Title",
                                        "max_length": 50,
                                    }
                                },
                            },
                        },
                    }
                },
            },
        )

    def test_options_on_serializer_with_exclude_kwarg(self):
        url = reverse_lazy("course_with_exclude_kwarg-list")
        response = self.client.options(url, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Course With Exclude Kwarg List",
                "description": "",
                "renders": ["application/json", "text/html"],
                "parses": [
                    "application/json",
                    "application/x-www-form-urlencoded",
                    "multipart/form-data",
                ],
                "actions": {
                    "POST": {
                        "name": {
                            "type": "string",
                            "required": True,
                            "read_only": False,
                            "label": "Name",
                            "max_length": 50,
                        },
                        "code": {
                            "type": "string",
                            "required": True,
                            "read_only": False,
                            "label": "Code",
                            "max_length": 30,
                        },
                        "books": {
                            "type": "field",
                            "required": False,
                            "read_only": True,
                            "label": "Books",
                            "child": {
                                "type": "nested object",
                                "required": False,
                                "read_only": True,
                                "children": {
                                    "title": {
                                        "type": "string",
                                        "required": True,
                                        "read_only": False,
                                        "label": "Title",
                                        "max_length": 50,
                                    }
                                },
                            },
                        },
                    }
                },
            },
        )


class DataMutationTests(APITestCase):
    def setUp(self):
        self.book1 = Book.objects.create(
            title="Advanced Data Structures", author="S.Mobit"
        )
        self.book2 = Book.objects.create(
            title="Basic Data Structures", author="S.Mobit"
        )

        self.instructor = Instructor.objects.create(name="Glady")

        self.course1 = Course.objects.create(name="Data Structures", code="CS210")
        self.course2 = Course.objects.create(name="Programming", code="CS150")

        self.course1.books.set([self.book1, self.book2])
        self.course2.books.set([self.book1])

        self.student = Student.objects.create(name="Yezy", age=24, course=self.course1)

        self.phone1 = Phone.objects.create(
            number="076711110", type="Office", student=self.student
        )
        self.phone2 = Phone.objects.create(
            number="073008880", type="Home", student=self.student
        )

    def tearDown(self):
        Book.objects.all().delete()
        Course.objects.all().delete()
        Student.objects.all().delete()

    # **************** Requestless Tests ********************* #

    def test_creating_data_without_request(self):
        serializer = WritableCourseSerializer(
            data={"name": "Data Structures", "code": "CS110", "books": {"add": [1, 2]}},
            query="{name, code, books{title}}",
        )

        serializer.is_valid()
        serializer.save()

        self.assertEqual(
            serializer.data,
            {
                "name": "Data Structures",
                "code": "CS110",
                "books": [
                    {"title": "Advanced Data Structures"},
                    {"title": "Basic Data Structures"},
                ],
            },
        )

    def test_updating_data_without_request(self):
        serializer = WritableStudentSerializer(
            self.student,
            data={
                "name": "yezy",
                "age": 33,
                "course": {
                    "name": "Programming",
                    "code": "CS50",
                    "books": {"remove": [1]},
                },
            },
            query="{name, age, course{name, books}, phone_numbers}",
        )

        serializer.is_valid()
        serializer.save()

        self.assertEqual(
            serializer.data,
            {
                "name": "yezy",
                "age": 33,
                "course": {
                    "name": "Programming",
                    "books": [
                        {
                            "title": "Basic Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        }
                    ],
                },
                "phone_numbers": [
                    {"number": "076711110", "type": "Office", "student": 1},
                    {"number": "073008880", "type": "Home", "student": 1},
                ],
            },
        )

    # **************** POST Tests ********************* #

    def test_post_on_pk_nested_foreignkey_related_field(self):
        url = reverse_lazy("rstudent-list")
        data = {"name": "yezy", "age": 33, "course": 2}
        response = self.client.post(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "yezy",
                "age": 33,
                "course": {
                    "name": "Programming",
                    "code": "CS150",
                    "books": [
                        {
                            "title": "Advanced Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        }
                    ],
                    "instructor": None,
                },
                "phone_numbers": [],
            },
        )

    def test_post_on_pk_writable_nested_foreignkey_related_field(self):
        url = reverse_lazy("rstudent-list")
        data = {
            "name": "yezy",
            "age": 33,
            "course": {"name": "Electronics", "code": "E320"},
        }
        response = self.client.post(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "yezy",
                "age": 33,
                "course": {
                    "name": "Electronics",
                    "code": "E320",
                    "books": [],
                    "instructor": None,
                },
                "phone_numbers": [],
            },
        )

    def test_post_on_pk_nested_foreignkey_related_field_with_alias(self):
        url = reverse_lazy("rstudent_with_alias-list")
        data = {"full_name": "yezy", "age": 33, "program": 2}
        response = self.client.post(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "full_name": "yezy",
                "age": 33,
                "program": {
                    "name": "Programming",
                    "code": "CS150",
                    "books": [
                        {
                            "title": "Advanced Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        }
                    ],
                    "instructor": None,
                },
                "contacts": [],
            },
        )

    def test_post_on_pk_nested_nullable_foreignkey_related_field_with_null_not_passed(
        self,
    ):
        url = reverse_lazy("rstudent-list")
        data = {"name": "yezy", "age": 33}
        response = self.client.post(url, data, format="json")

        self.assertEqual(
            response.data,
            {"name": "yezy", "age": 33, "course": None, "phone_numbers": []},
        )

    def test_post_on_pk_nested_nullable_foreignkey_related_field_with_null_passed(self):
        url = reverse_lazy("rstudent-list")
        data = {"name": "yezy", "age": 33, "course": None}
        response = self.client.post(url, data, format="json")

        self.assertEqual(
            response.data,
            {"name": "yezy", "age": 33, "course": None, "phone_numbers": []},
        )

    def test_post_on_writable_nested_foreignkey_related_field(self):
        url = reverse_lazy("wstudent-list")
        data = {
            "name": "yezy",
            "age": 33,
            "course": {"name": "Programming", "code": "CS50"},
        }
        response = self.client.post(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "yezy",
                "age": 33,
                "course": {
                    "name": "Programming",
                    "code": "CS50",
                    "books": [],
                    "instructor": None,
                },
                "phone_numbers": [],
            },
        )

    def test_post_on_writable_nested_foreignkey_related_field_with_aliased_fields(self):
        url = reverse_lazy("wstudent_with_alias-list")
        data = {
            "name": "yezy",
            "age": 27,
            "program": {"name": "Programming", "code": "CS50"},
            "contacts": {"add": [1]},
        }
        response = self.client.post(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "yezy",
                "age": 27,
                "program": {
                    "name": "Programming",
                    "code": "CS50",
                    "books": [],
                    "instructor": None,
                },
                "contacts": [{"number": "076711110", "type": "Office", "student": 2}],
                "study_partner": None,
                "sport_mates": [],
            },
        )

    def test_post_on_writable_nested_nullable_foreignkey_related_field(self):
        url = reverse_lazy("wstudent-list")
        data = {"name": "yezy", "age": 33, "course": None}
        response = self.client.post(url, data, format="json")

        self.assertEqual(
            response.data,
            {"name": "yezy", "age": 33, "course": None, "phone_numbers": []},
        )

    def test_post_with_add_operation(self):
        url = reverse_lazy("wcourse-list")
        data = {"name": "Data Structures", "code": "CS310", "books": {"add": [1, 2]}}
        response = self.client.post(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Data Structures",
                "code": "CS310",
                "books": [
                    {
                        "title": "Advanced Data Structures",
                        "author": "S.Mobit",
                        "genres": [],
                    },
                    {
                        "title": "Basic Data Structures",
                        "author": "S.Mobit",
                        "genres": [],
                    },
                ],
                "instructor": None,
            },
        )

    def test_post_with_create_operation(self):
        data = {
            "name": "Data Structures",
            "code": "CS310",
            "books": {
                "create": [
                    {"title": "Linear Math", "author": "Me"},
                    {"title": "Algebra Three", "author": "Me"},
                ]
            },
        }
        url = reverse_lazy("wcourse-list")
        response = self.client.post(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Data Structures",
                "code": "CS310",
                "books": [
                    {"title": "Linear Math", "author": "Me", "genres": []},
                    {"title": "Algebra Three", "author": "Me", "genres": []},
                ],
                "instructor": None,
            },
        )

    def test_post_with_nesting_under_create_operation(self):
        data = {
            "name": "Data Structures",
            "code": "CS315",
            "books": {
                "create": [
                    {
                        "title": "Linear Math 2",
                        "author": "B.Chris",
                        "genres": {
                            "create": [
                                {"title": "Science", "description": "All about nature"}
                            ]
                        },
                    },
                ]
            },
        }
        url = reverse_lazy("wcourse-list")
        response = self.client.post(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Data Structures",
                "code": "CS315",
                "books": [
                    {
                        "title": "Linear Math 2",
                        "author": "B.Chris",
                        "genres": [
                            {"title": "Science", "description": "All about nature"}
                        ],
                    },
                ],
                "instructor": None,
            },
        )

    def test_post_with_add_and_create_operations(self):
        data = {
            "name": "Data Structures",
            "code": "CS310",
            "books": {
                "add": [1],
                "create": [{"title": "Algebra Three", "author": "Me"}],
            },
        }
        url = reverse_lazy("wcourse-list")
        response = self.client.post(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Data Structures",
                "code": "CS310",
                "books": [
                    {
                        "title": "Advanced Data Structures",
                        "author": "S.Mobit",
                        "genres": [],
                    },
                    {"title": "Algebra Three", "author": "Me", "genres": []},
                ],
                "instructor": None,
            },
        )

    def test_post_on_deep_nested_fields(self):
        url = reverse_lazy("wstudent-list")
        data = {
            "name": "yezy",
            "age": 33,
            "course": {
                "name": "Programming",
                "code": "CS50",
                "books": {
                    "create": [
                        {
                            "title": "Python Tricks",
                            "author": "Dan Bader",
                            "genres": {
                                "create": [
                                    {
                                        "title": "Computing",
                                        "description": "Computer science",
                                    }
                                ]
                            },
                        }
                    ]
                },
            },
        }
        response = self.client.post(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "yezy",
                "age": 33,
                "course": {
                    "name": "Programming",
                    "code": "CS50",
                    "books": [
                        {
                            "title": "Python Tricks",
                            "author": "Dan Bader",
                            "genres": [
                                {
                                    "title": "Computing",
                                    "description": "Computer science",
                                }
                            ],
                        }
                    ],
                    "instructor": None,
                },
                "phone_numbers": [],
            },
        )

    def test_post_on_many_2_one_relation(self):
        url = reverse_lazy("wstudent-list")
        data = {
            "name": "yezy",
            "age": 33,
            "course": {"name": "Programming", "code": "CS50"},
            "phone_numbers": {"create": [{"number": "076750000", "type": "office"}]},
        }
        response = self.client.post(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "yezy",
                "age": 33,
                "course": {
                    "name": "Programming",
                    "code": "CS50",
                    "books": [],
                    "instructor": None,
                },
                "phone_numbers": [
                    {"number": "076750000", "type": "office", "student": 2}
                ],
            },
        )

    def test_post_with_pk_field_inside_writable_nested_field(self):
        url = reverse_lazy("wstudent-list")
        data = {
            "name": "Gray",
            "age": 28,
            "course": {
                "name": "Programming",
                "code": "CS50",
                "instructor": self.instructor.id,
            },
        }
        response = self.client.post(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Gray",
                "age": 28,
                "course": {
                    "name": "Programming",
                    "code": "CS50",
                    "books": [],
                    "instructor": {"name": "Glady"},
                },
                "phone_numbers": [],
            },
        )

    # **************** POST with generic relations ********************* #
    def test_post_on_generic_relation_field(self):
        existing_attachment = Attachment.objects.create(
            content="Existing attachment", document=Post.objects.create()
        )

        response = self.client.post(
            reverse_lazy("post-list"),
            {
                "title": "New post",
                "content": "This is the post",
                "attachments": {
                    "create": [
                        {"content": "This is first attachment"},
                        {"content": "This is second attachment"},
                    ],
                    "add": [existing_attachment.pk],
                },
            },
            format="json",
        )
        self.assertEqual(response.status_code, 201)
        self.assertEqual(Post.objects.count(), 2)
        post = Post.objects.get(title="New post")
        self.assertEqual(Attachment.objects.filter(post=post).count(), 3)
        attachment = Attachment.objects.last()
        self.assertEqual(attachment.content, "This is second attachment")
        existing_attachment.refresh_from_db()
        self.assertEqual(existing_attachment.document, post)

    def test_post_without_attachments(self):
        response = self.client.post(
            reverse_lazy("post-list"),
            data={"title": "Olympic games", "content": "Test post without attachments"},
            format="json",
        )
        self.assertEqual(response.status_code, 201)
        self.assertEqual(Post.objects.count(), 1)
        self.assertEqual(Post.objects.first().attachments.count(), 0)

    # **************** PUT Tests ********************* #

    def test_put_on_pk_nested_foreignkey_related_field(self):
        url = reverse_lazy("rstudent-detail", args=[self.student.id])
        data = {"name": "yezy", "age": 33, "course": 2}
        response = self.client.put(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "yezy",
                "age": 33,
                "course": {
                    "name": "Programming",
                    "code": "CS150",
                    "books": [
                        {
                            "title": "Advanced Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        }
                    ],
                    "instructor": None,
                },
                "phone_numbers": [
                    {"number": "076711110", "type": "Office", "student": 1},
                    {"number": "073008880", "type": "Home", "student": 1},
                ],
            },
        )

    def test_put_on_pk_nested_foreignkey_related_field_with_alias(self):
        url = reverse_lazy("rstudent_with_alias-detail", args=[self.student.id])
        data = {"full_name": "yezy", "age": 33, "program": 2}
        response = self.client.put(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "full_name": "yezy",
                "age": 33,
                "program": {
                    "name": "Programming",
                    "code": "CS150",
                    "books": [
                        {
                            "title": "Advanced Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        }
                    ],
                    "instructor": None,
                },
                "contacts": [
                    {"number": "076711110", "type": "Office", "student": 1},
                    {"number": "073008880", "type": "Home", "student": 1},
                ],
            },
        )

    def test_put_on_writable_nested_foreignkey_related_field(self):
        url = reverse_lazy("wstudent-detail", args=[self.student.id])
        data = {
            "name": "yezy",
            "age": 33,
            "course": {"name": "Programming", "code": "CS50"},
        }
        response = self.client.put(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "yezy",
                "age": 33,
                "course": {
                    "name": "Programming",
                    "code": "CS50",
                    "books": [
                        {
                            "title": "Advanced Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                        {
                            "title": "Basic Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                    ],
                    "instructor": None,
                },
                "phone_numbers": [
                    {"number": "076711110", "type": "Office", "student": 1},
                    {"number": "073008880", "type": "Home", "student": 1},
                ],
            },
        )

    def test_put_on_writable_nested_foreignkey_related_field_with_aliased_fields(self):
        url = reverse_lazy("wstudent_with_alias-detail", args=[self.student.id])
        data = {
            "name": "yezy",
            "age": 27,
            "program": {"name": "Programming & Data Analysis", "code": "CS55"},
            "contacts": {},
        }
        response = self.client.put(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "yezy",
                "age": 27,
                "program": {
                    "name": "Programming & Data Analysis",
                    "code": "CS55",
                    "books": [
                        {
                            "title": "Advanced Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                        {
                            "title": "Basic Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                    ],
                    "instructor": None,
                },
                "contacts": [
                    {"number": "076711110", "type": "Office", "student": 1},
                    {"number": "073008880", "type": "Home", "student": 1},
                ],
                "study_partner": None,
                "sport_mates": [],
            },
        )

    def test_put_on_writable_nested_nullable_foreignkey_related_field(self):
        url = reverse_lazy("wstudent-detail", args=[self.student.id])
        data = {"name": "yezy", "age": 33}
        response = self.client.put(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "yezy",
                "age": 33,
                "course": {
                    "name": "Data Structures",
                    "code": "CS210",
                    "books": [
                        {
                            "title": "Advanced Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                        {
                            "title": "Basic Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                    ],
                    "instructor": None,
                },
                "phone_numbers": [
                    {"number": "076711110", "type": "Office", "student": 1},
                    {"number": "073008880", "type": "Home", "student": 1},
                ],
            },
        )

    def test_put_on_writable_nested_nullable_foreignkey_related_field_with_empty_string(
        self,
    ):
        url = reverse_lazy("wstudent-detail", args=[self.student.id])
        data = {"name": "yezy", "age": 33, "course": None}
        response = self.client.put(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "yezy",
                "age": 33,
                "course": None,
                "phone_numbers": [
                    {"number": "076711110", "type": "Office", "student": 1},
                    {"number": "073008880", "type": "Home", "student": 1},
                ],
            },
        )

    def test_put_with_add_operation(self):
        url = reverse_lazy("wcourse-detail", args=[self.course2.id])
        data = {"name": "Data Structures", "code": "CS410", "books": {"add": [2]}}
        response = self.client.put(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Data Structures",
                "code": "CS410",
                "books": [
                    {
                        "title": "Advanced Data Structures",
                        "author": "S.Mobit",
                        "genres": [],
                    },
                    {
                        "title": "Basic Data Structures",
                        "author": "S.Mobit",
                        "genres": [],
                    },
                ],
                "instructor": None,
            },
        )

    def test_put_with_remove_operation(self):
        url = reverse_lazy("wcourse-detail", args=[self.course2.id])
        data = {"name": "Data Structures", "code": "CS410", "books": {"remove": [1]}}
        response = self.client.put(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Data Structures",
                "code": "CS410",
                "books": [],
                "instructor": None,
            },
        )

    def test_put_with_create_operation(self):
        url = reverse_lazy("wcourse-detail", args=[self.course2.id])
        data = {
            "name": "Data Structures",
            "code": "CS310",
            "books": {
                "create": [{"title": "Primitive Data Types", "author": "S.Mobit"}]
            },
        }
        response = self.client.put(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Data Structures",
                "code": "CS310",
                "books": [
                    {
                        "title": "Advanced Data Structures",
                        "author": "S.Mobit",
                        "genres": [],
                    },
                    {
                        "title": "Primitive Data Types",
                        "author": "S.Mobit",
                        "genres": [],
                    },
                ],
                "instructor": None,
            },
        )

    def test_put_with_update_operation(self):
        url = reverse_lazy("wcourse-detail", args=[self.course2.id])
        data = {
            "name": "Data Structures",
            "code": "CS310",
            "books": {
                "update": {1: {"title": "React Programming", "author": "M.Json"}}
            },
        }
        response = self.client.put(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Data Structures",
                "code": "CS310",
                "books": [
                    {"title": "React Programming", "author": "M.Json", "genres": []}
                ],
                "instructor": None,
            },
        )

    def test_put_with_update_operation_missing_one_required_nested_field(self):
        url = reverse_lazy("wcourse-detail", args=[self.course2.id])
        data = {
            "name": "Data Structures",
            "code": "CS310",
            "books": {
                "update": {
                    1: {"title": "React Programming"}  # We've removed author field
                }
            },
        }
        response = self.client.put(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Data Structures",
                "code": "CS310",
                "books": [
                    {"title": "React Programming", "author": "S.Mobit", "genres": []}
                ],
                "instructor": None,
            },
        )

    def test_put_on_deep_nested_fields(self):
        url = reverse_lazy("wstudent-detail", args=[self.student.id])
        data = {
            "name": "yezy",
            "age": 33,
            "course": {"name": "Programming", "code": "CS50", "books": {"remove": [1]}},
        }
        response = self.client.put(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "yezy",
                "age": 33,
                "course": {
                    "name": "Programming",
                    "code": "CS50",
                    "books": [
                        {
                            "title": "Basic Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        }
                    ],
                    "instructor": None,
                },
                "phone_numbers": [
                    {"number": "076711110", "type": "Office", "student": 1},
                    {"number": "073008880", "type": "Home", "student": 1},
                ],
            },
        )

    def test_put_on_many_2_one_relation(self):
        url = reverse_lazy("wstudent-detail", args=[self.student.id])
        data = {
            "name": "yezy",
            "age": 33,
            "course": {"name": "Programming", "code": "CS50"},
            "phone_numbers": {
                "update": {1: {"number": "073008811", "type": "office"}},
                "create": [{"number": "076750000", "type": "office"}],
            },
        }
        response = self.client.put(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "yezy",
                "age": 33,
                "course": {
                    "name": "Programming",
                    "code": "CS50",
                    "books": [
                        {
                            "title": "Advanced Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                        {
                            "title": "Basic Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                    ],
                    "instructor": None,
                },
                "phone_numbers": [
                    {"number": "073008811", "type": "office", "student": 1},
                    {"number": "073008880", "type": "Home", "student": 1},
                    {"number": "076750000", "type": "office", "student": 1},
                ],
            },
        )

    def test_put_with__all__as_remove_operation_value(self):
        url = reverse_lazy("wstudent-detail", args=[self.student.id])
        data = {
            "name": "Davinci",
            "age": 23,
            "course": {
                "name": "Programming",
                "code": "CS50",
                "books": {"remove": "__all__"},
            },
            "phone_numbers": {"remove": "__all__"},
        }
        response = self.client.put(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Davinci",
                "age": 23,
                "course": {
                    "name": "Programming",
                    "code": "CS50",
                    "books": [],
                    "instructor": None,
                },
                "phone_numbers": [],
            },
        )

    # **************** PATCH Tests ********************* #

    def test_patch_on_pk_nested_foreignkey_related_field(self):
        url = reverse_lazy("rstudent-detail", args=[self.student.id])
        data = {"course": 2}
        response = self.client.patch(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Yezy",
                "age": 24,
                "course": {
                    "name": "Programming",
                    "code": "CS150",
                    "books": [
                        {
                            "title": "Advanced Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        }
                    ],
                    "instructor": None,
                },
                "phone_numbers": [
                    {"number": "076711110", "type": "Office", "student": 1},
                    {"number": "073008880", "type": "Home", "student": 1},
                ],
            },
        )

    def test_patch_without_pk_nested_foreignkey_related_field(self):
        url = reverse_lazy("rstudent-detail", args=[self.student.id])
        data = {
            "age": 35,
        }
        response = self.client.patch(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Yezy",
                "age": 35,
                "course": {
                    "name": "Data Structures",
                    "code": "CS210",
                    "books": [
                        {
                            "title": "Advanced Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                        {
                            "title": "Basic Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                    ],
                    "instructor": None,
                },
                "phone_numbers": [
                    {"number": "076711110", "type": "Office", "student": 1},
                    {"number": "073008880", "type": "Home", "student": 1},
                ],
            },
        )

    def test_patch_on_pk_nested_foreignkey_related_field_with_alias(self):
        url = reverse_lazy("rstudent_with_alias-detail", args=[self.student.id])
        data = {"age": 33, "program": 2}
        response = self.client.patch(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "full_name": "Yezy",
                "age": 33,
                "program": {
                    "name": "Programming",
                    "code": "CS150",
                    "books": [
                        {
                            "title": "Advanced Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        }
                    ],
                    "instructor": None,
                },
                "contacts": [
                    {"number": "076711110", "type": "Office", "student": 1},
                    {"number": "073008880", "type": "Home", "student": 1},
                ],
            },
        )

    def test_patch_on_pk_nested_nullable_foreignkey_related_field(self):
        url = reverse_lazy("rstudent-detail", args=[self.student.id])
        data = {"age": 30}
        response = self.client.patch(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Yezy",
                "age": 30,
                "course": {
                    "name": "Data Structures",
                    "code": "CS210",
                    "books": [
                        {
                            "title": "Advanced Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                        {
                            "title": "Basic Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                    ],
                    "instructor": None,
                },
                "phone_numbers": [
                    {"number": "076711110", "type": "Office", "student": 1},
                    {"number": "073008880", "type": "Home", "student": 1},
                ],
            },
        )

    def test_patch_on_writable_nested_foreignkey_related_field(self):
        url = reverse_lazy("wstudent-detail", args=[self.student.id])
        data = {"name": "Yezy Ilomo", "course": {"name": "Programming", "code": "CS50"}}
        response = self.client.patch(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Yezy Ilomo",
                "age": 24,
                "course": {
                    "name": "Programming",
                    "code": "CS50",
                    "books": [
                        {
                            "title": "Advanced Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                        {
                            "title": "Basic Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                    ],
                    "instructor": None,
                },
                "phone_numbers": [
                    {"number": "076711110", "type": "Office", "student": 1},
                    {"number": "073008880", "type": "Home", "student": 1},
                ],
            },
        )

    def test_patch_on_writable_nested_foreignkey_related_field_with_aliased_fields(
        self,
    ):
        url = reverse_lazy("wstudent_with_alias-detail", args=[self.student.id])
        data = {
            "age": 28,
            "program": {"name": "Programming & Data Analysis", "code": "CS55"},
            "contacts": {},
        }
        response = self.client.patch(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Yezy",
                "age": 28,
                "program": {
                    "name": "Programming & Data Analysis",
                    "code": "CS55",
                    "books": [
                        {
                            "title": "Advanced Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                        {
                            "title": "Basic Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                    ],
                    "instructor": None,
                },
                "contacts": [
                    {"number": "076711110", "type": "Office", "student": 1},
                    {"number": "073008880", "type": "Home", "student": 1},
                ],
                "study_partner": None,
                "sport_mates": [],
            },
        )

    def test_patch_on_writable_nested_nullable_foreignkey_related_field_without_null_value(
        self,
    ):
        url = reverse_lazy("wstudent-detail", args=[self.student.id])
        data = {
            "name": "yezy",
            "age": 26,
            # "course": None  # We are not passing null
        }
        response = self.client.patch(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "yezy",
                "age": 26,
                "course": {
                    "name": "Data Structures",
                    "code": "CS210",
                    "books": [
                        {
                            "title": "Advanced Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                        {
                            "title": "Basic Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                    ],
                    "instructor": None,
                },
                "phone_numbers": [
                    {"number": "076711110", "type": "Office", "student": 1},
                    {"number": "073008880", "type": "Home", "student": 1},
                ],
            },
        )

    def test_patch_on_writable_nested_nullable_foreignkey_related_field_with_null_value(
        self,
    ):
        url = reverse_lazy("wstudent-detail", args=[self.student.id])
        data = {"name": "yezy", "age": 33, "course": None}
        response = self.client.patch(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "yezy",
                "age": 33,
                "course": None,
                "phone_numbers": [
                    {"number": "076711110", "type": "Office", "student": 1},
                    {"number": "073008880", "type": "Home", "student": 1},
                ],
            },
        )

    def test_patch_with_add_operation(self):
        url = reverse_lazy("wcourse-detail", args=[self.course2.id])
        data = {"name": "Data Structures", "code": "CS410", "books": {"add": [2]}}
        response = self.client.patch(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Data Structures",
                "code": "CS410",
                "books": [
                    {
                        "title": "Advanced Data Structures",
                        "author": "S.Mobit",
                        "genres": [],
                    },
                    {
                        "title": "Basic Data Structures",
                        "author": "S.Mobit",
                        "genres": [],
                    },
                ],
                "instructor": None,
            },
        )

    def test_patch_with_remove_operation(self):
        url = reverse_lazy("wcourse-detail", args=[self.course2.id])
        data = {"name": "Data Structures", "code": "CS410", "books": {"remove": [1]}}
        response = self.client.patch(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Data Structures",
                "code": "CS410",
                "books": [],
                "instructor": None,
            },
        )

    def test_patch_with_create_operation(self):
        url = reverse_lazy("wcourse-detail", args=[self.course2.id])
        data = {
            "name": "Data Structures",
            "code": "CS310",
            "books": {
                "create": [{"title": "Primitive Data Types", "author": "S.Mobit"}]
            },
        }
        response = self.client.patch(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Data Structures",
                "code": "CS310",
                "books": [
                    {
                        "title": "Advanced Data Structures",
                        "author": "S.Mobit",
                        "genres": [],
                    },
                    {
                        "title": "Primitive Data Types",
                        "author": "S.Mobit",
                        "genres": [],
                    },
                ],
                "instructor": None,
            },
        )

    def test_patch_with_update_operation(self):
        url = reverse_lazy("wcourse-detail", args=[self.course2.id])
        data = {
            "name": "Data Structures",
            "code": "CS310",
            "books": {
                "update": {1: {"title": "React Programming", "author": "M.Json"}}
            },
        }
        response = self.client.patch(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Data Structures",
                "code": "CS310",
                "books": [
                    {"title": "React Programming", "author": "M.Json", "genres": []}
                ],
                "instructor": None,
            },
        )

    def test_patch_with_update_operation_missing_one_required_nested_field(self):
        url = reverse_lazy("wcourse-detail", args=[self.course2.id])
        data = {
            "name": "Data Structures",
            "code": "CS310",
            "books": {
                "update": {
                    1: {"title": "React Programming"}  # We've removed author field
                }
            },
        }
        response = self.client.patch(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Data Structures",
                "code": "CS310",
                "books": [
                    {"title": "React Programming", "author": "S.Mobit", "genres": []}
                ],
                "instructor": None,
            },
        )

    def test_patch_with_mixed_operations(self):
        url = reverse_lazy("wstudent-detail", args=[self.student.id])
        data = {
            "name": "yezy",
            "age": 33,
            "course": {
                "name": "Programming",
                "code": "CS50",
                "books": {
                    "update": {
                        1: {
                            "title": "React Programming",
                            "author": "K.Kennedy",
                            "genres": {
                                # update_operations applies here since the parent is "update"
                                "remove": [],
                                "add": [],
                                "create": [
                                    {
                                        "title": "Modern Programming",
                                        "description": "New tools",
                                    }
                                ],
                                "update": {},
                            },
                        }
                    },
                    "create": [
                        {
                            "title": "CS Foundation",
                            "author": "T.Howard",
                            "genres": {
                                # create_operations applies here since the parent is "create"
                                "add": [],
                                "create": [
                                    {
                                        "title": "Classical Programming",
                                        "description": "Classical tech tools",
                                    }
                                ],
                            },
                        }
                    ],
                },
            },
        }
        response = self.client.patch(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "yezy",
                "age": 33,
                "course": {
                    "name": "Programming",
                    "code": "CS50",
                    "books": [
                        {
                            "title": "React Programming",
                            "author": "K.Kennedy",
                            "genres": [
                                {
                                    "title": "Modern Programming",
                                    "description": "New tools",
                                }
                            ],
                        },
                        {
                            "title": "Basic Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                        {
                            "title": "CS Foundation",
                            "author": "T.Howard",
                            "genres": [
                                {
                                    "title": "Classical Programming",
                                    "description": "Classical tech tools",
                                }
                            ],
                        },
                    ],
                    "instructor": None,
                },
                "phone_numbers": [
                    {"number": "076711110", "type": "Office", "student": 1},
                    {"number": "073008880", "type": "Home", "student": 1},
                ],
            },
        )

    def test_patch_on_many_2_one_relation(self):
        url = reverse_lazy("wstudent-detail", args=[self.student.id])
        data = {
            "name": "yezy",
            "age": 33,
            "course": {"name": "Programming", "code": "CS50"},
            "phone_numbers": {
                "update": {1: {"number": "073008811", "type": "office"}},
                "create": [{"number": "076750000", "type": "office"}],
            },
        }
        response = self.client.patch(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "yezy",
                "age": 33,
                "course": {
                    "name": "Programming",
                    "code": "CS50",
                    "books": [
                        {
                            "title": "Advanced Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                        {
                            "title": "Basic Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                    ],
                    "instructor": None,
                },
                "phone_numbers": [
                    {"number": "073008811", "type": "office", "student": 1},
                    {"number": "073008880", "type": "Home", "student": 1},
                    {"number": "076750000", "type": "office", "student": 1},
                ],
            },
        )

    def test_patch_with__all__as_remove_operation_value(self):
        url = reverse_lazy("wstudent-detail", args=[self.student.id])
        data = {
            "name": "Davinci",
            "age": 33,
            "course": {
                "name": "Programming",
                "code": "CS50",
                "books": {"remove": "__all__"},
            },
            "phone_numbers": {"remove": "__all__"},
        }
        response = self.client.patch(url, data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Davinci",
                "age": 33,
                "course": {
                    "name": "Programming",
                    "code": "CS50",
                    "books": [],
                    "instructor": None,
                },
                "phone_numbers": [],
            },
        )

    def test_patch_with_mixed_operations_on_generic_relation(self):
        post = Post.objects.create()
        attachment = Attachment.objects.create(document=post)
        removed_attachment = Attachment.objects.create(document=post, content="removed")
        attachment_to_add = Attachment.objects.create(
            content="attachment to add",
            document=Post.objects.create(title="Another post"),
        )

        response = self.client.patch(
            reverse_lazy("post-detail", args=[post.pk]),
            {
                "attachments": {
                    "create": [
                        {"content": "new attachment"},
                    ],
                    "update": {str(attachment.pk): {"content": "old attachment"}},
                    "remove": [removed_attachment.pk],
                    "add": [attachment_to_add.pk],
                }
            },
            format="json",
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(Attachment.objects.filter(post=post).count(), 3)
        attachment.refresh_from_db()
        self.assertEqual(attachment.content, "old attachment")
        with self.assertRaises(Attachment.DoesNotExist):
            removed_attachment.refresh_from_db()

        attachment_to_add.refresh_from_db()
        self.assertEqual(attachment_to_add.document, post)
        self.assertEqual(
            Attachment.objects.get(content="new attachment").document, post
        )


class DataQueryingAndMutationTests(APITestCase):
    def setUp(self):
        self.book1 = Book.objects.create(
            title="Advanced Data Structures", author="S.Mobit"
        )
        self.book2 = Book.objects.create(
            title="Basic Data Structures", author="S.Mobit"
        )

        self.course1 = Course.objects.create(name="Data Structures", code="CS210")
        self.course2 = Course.objects.create(name="Programming", code="CS150")

        self.course1.books.set([self.book1, self.book2])
        self.course2.books.set([self.book1])

        self.student = Student.objects.create(name="Yezy", age=24, course=self.course1)

        self.phone1 = Phone.objects.create(
            number="076711110", type="Office", student=self.student
        )
        self.phone2 = Phone.objects.create(
            number="073008880", type="Home", student=self.student
        )

    def tearDown(self):
        Book.objects.all().delete()
        Course.objects.all().delete()
        Student.objects.all().delete()

    # **************** POST Tests ********************* #

    def test_post_on_pk_nested_foreignkey_related_field_mix_with_query_param(self):
        url = reverse_lazy("rstudent-list")
        data = {"name": "yezy", "age": 33, "course": 2}
        response = self.client.post(url + "?query={course}", data, format="json")

        self.assertEqual(
            response.data,
            {
                "course": {
                    "name": "Programming",
                    "code": "CS150",
                    "books": [
                        {
                            "title": "Advanced Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        }
                    ],
                    "instructor": None,
                }
            },
        )

    def test_post_on_pk_nested_foreignkey_related_field_with_alias_mix_with_query_param(
        self,
    ):
        url = reverse_lazy("rstudent_with_alias-list")
        data = {"full_name": "yezy", "age": 33, "program": 2}
        response = self.client.post(
            url + "?query={full_name, program}", data, format="json"
        )

        self.assertEqual(
            response.data,
            {
                "full_name": "yezy",
                "program": {
                    "name": "Programming",
                    "code": "CS150",
                    "books": [
                        {
                            "title": "Advanced Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        }
                    ],
                    "instructor": None,
                },
            },
        )

    def test_post_on_writable_nested_foreignkey_related_field_mix_with_query_param(
        self,
    ):
        url = reverse_lazy("wstudent-list")
        data = {
            "name": "yezy",
            "age": 33,
            "course": {"name": "Programming", "code": "CS50"},
        }
        response = self.client.post(
            url + "?query={course{name, books{}}}", data, format="json"
        )

        self.assertEqual(
            response.data, {"course": {"name": "Programming", "books": []}}
        )

    def test_post_with_add_operation_mix_with_query_param(self):
        url = reverse_lazy("wcourse-list")
        data = {"name": "Data Structures", "code": "CS310", "books": {"add": [1, 2]}}
        response = self.client.post(url + "?query={books{title}}", data, format="json")

        self.assertEqual(
            response.data,
            {
                "books": [
                    {"title": "Advanced Data Structures"},
                    {"title": "Basic Data Structures"},
                ]
            },
        )

    def test_post_with_add_operation_mix_with_query_param_and_alias_fields(self):
        url = reverse_lazy("wcourse-list")
        data = {"name": "Data Structures", "code": "CS310", "books": {"add": [1, 2]}}
        response = self.client.post(
            url + "?query={books{bookTitle: title}}", data, format="json"
        )

        self.assertEqual(
            response.data,
            {
                "books": [
                    {"bookTitle": "Advanced Data Structures"},
                    {"bookTitle": "Basic Data Structures"},
                ]
            },
        )

    def test_post_on_self_referencing_nested_field(self):
        url = reverse_lazy("wstudent_with_alias-list")
        data = {
            "name": "Gwen",
            "age": 27,
            "program": {"name": "Programming & Data Analysis", "code": "CS010"},
            "contacts": {},
            "study_partner": {"name": "Ilomo", "age": 24},
            "sport_mates": {
                "add": [1],
                "create": [{"name": "Yoram", "age": 26}, {"name": "Gaby", "age": 23}],
            },
        }
        response = self.client.post(
            url + "?query={*, study_partner{name}, sport_mates{name, age}}",
            data,
            format="json",
        )

        self.assertEqual(
            response.data,
            {
                "name": "Gwen",
                "age": 27,
                "program": {
                    "name": "Programming & Data Analysis",
                    "code": "CS010",
                    "books": [],
                    "instructor": None,
                },
                "contacts": [],
                "study_partner": {"name": "Ilomo"},
                "sport_mates": [
                    {"name": "Yezy", "age": 24},
                    {"name": "Yoram", "age": 26},
                    {"name": "Gaby", "age": 23},
                ],
            },
        )

    # **************** PUT Tests ********************* #

    def test_put_on_pk_nested_foreignkey_related_field_mix_with_query_param(self):
        url = reverse_lazy("rstudent-detail", args=[self.student.id])
        data = {"name": "Yezy", "age": 25, "course": 2}
        response = self.client.put(url + "?query={course}", data, format="json")

        self.assertEqual(
            response.data,
            {
                "course": {
                    "name": "Programming",
                    "code": "CS150",
                    "books": [
                        {
                            "title": "Advanced Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        }
                    ],
                    "instructor": None,
                }
            },
        )

    def test_put_on_writable_nested_foreignkey_related_field_mix_with_query_param(self):
        url = reverse_lazy("wstudent-detail", args=[self.student.id])
        data = {
            "name": "Yezy",
            "age": 25,
            "course": {"name": "Programming", "code": "CS50"},
        }
        response = self.client.put(url + "?query={name, course}", data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Yezy",
                "course": {
                    "name": "Programming",
                    "code": "CS50",
                    "books": [
                        {
                            "title": "Advanced Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                        {
                            "title": "Basic Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                    ],
                    "instructor": None,
                },
            },
        )

    def test_put_on_deep_nested_fields_mix_with_query_param(self):
        url = reverse_lazy("wstudent-detail", args=[self.student.id])
        data = {
            "name": "yezy",
            "age": 33,
            "course": {"name": "Programming", "code": "CS50", "books": {"remove": [1]}},
        }
        response = self.client.put(
            url + "?query={course{books{title}}}", data, format="json"
        )

        self.assertEqual(
            response.data, {"course": {"books": [{"title": "Basic Data Structures"}]}}
        )

    def test_put_on_deep_nested_fields_mix_with_query_param_and_aliased_fields(self):
        url = reverse_lazy("wstudent-detail", args=[self.student.id])
        data = {
            "name": "yezy",
            "age": 33,
            "course": {"name": "Programming", "code": "CS50", "books": {"remove": [1]}},
        }
        response = self.client.put(
            url + "?query={programme: course{readings: books{bookTitle: title}}}",
            data,
            format="json",
        )

        self.assertEqual(
            response.data,
            {"programme": {"readings": [{"bookTitle": "Basic Data Structures"}]}},
        )

    def test_put_on_self_referencing_nested_field(self):
        url = reverse_lazy("wstudent_with_alias-detail", args=[self.student.id])
        data = {
            "name": "yezy",
            "age": 27,
            "program": {"name": "Programming & Data Analysis", "code": "CS55"},
            "contacts": {},
            "study_partner": {"name": "Ilomo", "age": 24},
            "sport_mates": {
                "create": [{"name": "Yoram", "age": 26}, {"name": "Gaby", "age": 23}]
            },
        }
        response = self.client.put(
            url + "?query={*, study_partner{name}, sport_mates{name, age}}",
            data,
            format="json",
        )

        self.assertEqual(
            response.data,
            {
                "name": "yezy",
                "age": 27,
                "program": {
                    "name": "Programming & Data Analysis",
                    "code": "CS55",
                    "books": [
                        {
                            "title": "Advanced Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                        {
                            "title": "Basic Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                    ],
                    "instructor": None,
                },
                "contacts": [
                    {"number": "076711110", "type": "Office", "student": 1},
                    {"number": "073008880", "type": "Home", "student": 1},
                ],
                "study_partner": {"name": "Ilomo"},
                "sport_mates": [
                    {"name": "Yoram", "age": 26},
                    {"name": "Gaby", "age": 23},
                ],
            },
        )

    # **************** PATCH Tests ********************* #

    def test_patch_on_pk_nested_foreignkey_related_field_mix_with_query_param(self):
        url = reverse_lazy("rstudent-detail", args=[self.student.id])
        data = {"course": 2}
        response = self.client.patch(url + "?query={name, course}", data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Yezy",
                "course": {
                    "name": "Programming",
                    "code": "CS150",
                    "books": [
                        {
                            "title": "Advanced Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        }
                    ],
                    "instructor": None,
                },
            },
        )

    def test_patch_on_writable_nested_foreignkey_related_field_mix_with_query_param(
        self,
    ):
        url = reverse_lazy("wstudent-detail", args=[self.student.id])
        data = {"name": "Yezy Ilomo", "course": {"name": "Programming", "code": "CS50"}}
        response = self.client.patch(url + "?query={name, course}", data, format="json")

        self.assertEqual(
            response.data,
            {
                "name": "Yezy Ilomo",
                "course": {
                    "name": "Programming",
                    "code": "CS50",
                    "books": [
                        {
                            "title": "Advanced Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                        {
                            "title": "Basic Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                    ],
                    "instructor": None,
                },
            },
        )

    def test_patch_on_deep_nested_fields_mix_with_query_param(self):
        url = reverse_lazy("wstudent-detail", args=[self.student.id])
        data = {
            "name": "yezy",
            "age": 33,
            "course": {"name": "Programming", "code": "CS50", "books": {"remove": [1]}},
        }
        response = self.client.patch(
            url + "?query={course{books{title}}}", data, format="json"
        )

        self.assertEqual(
            response.data, {"course": {"books": [{"title": "Basic Data Structures"}]}}
        )

    def test_patch_on_deep_nested_fields_mix_with_query_param_and_aliased_fields(self):
        url = reverse_lazy("wstudent-detail", args=[self.student.id])
        data = {
            "name": "yezy",
            "age": 33,
            "course": {"name": "Programming", "code": "CS50", "books": {"remove": [1]}},
        }
        response = self.client.patch(
            url + "?query={programme: course{books{name: title}}}", data, format="json"
        )

        self.assertEqual(
            response.data, {"programme": {"books": [{"name": "Basic Data Structures"}]}}
        )

    def test_patch_on_self_referencing_nested_field(self):
        url = reverse_lazy("wstudent_with_alias-detail", args=[self.student.id])
        data = {
            "name": "yezy",
            "age": 27,
            "program": {"name": "Programming & Data Analysis", "code": "CS55"},
            "contacts": {},
            "study_partner": 1,
            "sport_mates": {
                "create": [{"name": "Yoram", "age": 26}, {"name": "Gaby", "age": 23}]
            },
        }
        response = self.client.patch(
            url + "?query={*, study_partner{name}, sport_mates{name, age}}",
            data,
            format="json",
        )

        self.assertEqual(
            response.data,
            {
                "name": "yezy",
                "age": 27,
                "program": {
                    "name": "Programming & Data Analysis",
                    "code": "CS55",
                    "books": [
                        {
                            "title": "Advanced Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                        {
                            "title": "Basic Data Structures",
                            "author": "S.Mobit",
                            "genres": [],
                        },
                    ],
                    "instructor": None,
                },
                "contacts": [
                    {"number": "076711110", "type": "Office", "student": 1},
                    {"number": "073008880", "type": "Home", "student": 1},
                ],
                "study_partner": {"name": "Yezy"},
                "sport_mates": [
                    {"name": "Yoram", "age": 26},
                    {"name": "Gaby", "age": 23},
                ],
            },
        )



from rest_framework.serializers import ModelSerializer

from .mixins import NestedCreateMixin, NestedUpdateMixin


class NestedModelSerializer(
        NestedCreateMixin,
        NestedUpdateMixin,
        ModelSerializer):
    pass



class DjangoRESTQLException(Exception):
    """Base class for exceptions in this package."""


class InvalidOperation(DjangoRESTQLException):
    """Invalid Operation Exception."""


class FieldNotFound(DjangoRESTQLException):
    """Field Not Found Exception."""


class QueryFormatError(DjangoRESTQLException):
    """Invalid Query Format."""



from django.http import QueryDict
from django.db.models import Prefetch
from django.utils.functional import cached_property
from django.core.exceptions import ObjectDoesNotExist
from django.db.models.fields.related import ManyToManyRel, ManyToOneRel

try:
    from django.contrib.contenttypes.fields import GenericRel
    from django.contrib.contenttypes.models import ContentType
except ImportError:
    GenericRel = None
    ContentType = None

from rest_framework.serializers import ListSerializer, Serializer, ValidationError

from .exceptions import FieldNotFound, QueryFormatError
from .fields import (
    ALL_RELATED_OBJS,
    BaseRESTQLNestedField,
    DynamicSerializerMethodField,
    TemporaryNestedField,
)
from .operations import ADD, CREATE, REMOVE, UPDATE
from .parser import Query, QueryParser
from .settings import restql_settings


class RequestQueryParserMixin(object):
    """
    Mixin for parsing restql query from request.

    NOTE: We are using `request.GET` instead of
    `request.query_params` because this might be
    called before DRF request is created(i.e from dispatch).
    This means `request.query_params` might not be available
    when this mixin is used.
    """

    @classmethod
    def has_restql_query_param(cls, request):
        query_param_name = restql_settings.QUERY_PARAM_NAME
        return query_param_name in request.GET

    @classmethod
    def get_parsed_restql_query_from_req(cls, request):
        if hasattr(request, "parsed_restql_query"):
            # Use cached parsed restql query
            return request.parsed_restql_query
        raw_query = request.GET[restql_settings.QUERY_PARAM_NAME]
        parser = QueryParser()
        parsed_restql_query = parser.parse(raw_query)

        # Save parsed restql query to the request so that
        # we won't need to parse it again if needed later
        request.parsed_restql_query = parsed_restql_query
        return parsed_restql_query


class QueryArgumentsMixin(RequestQueryParserMixin):
    """Mixin for converting query arguments into query parameters"""

    def get_parsed_restql_query(self, request):
        if self.has_restql_query_param(request):
            try:
                return self.get_parsed_restql_query_from_req(request)
            except (SyntaxError, QueryFormatError):
                # Let `DynamicFieldsMixin` handle this for a user
                # to get a helpful error message
                pass

        # Else include all fields
        query = Query(
            field_name=None,
            included_fields=["*"],
            excluded_fields=[],
            aliases={},
            arguments={},
        )
        return query

    def build_query_params(self, parsed_query, parent=None):
        query_params = {}
        prefix = ""
        if parent is None:
            query_params.update(parsed_query.arguments)
        else:
            prefix = parent + "__"
            for argument, value in parsed_query.arguments.items():
                name = prefix + argument
                query_params.update({name: value})

        for field in parsed_query.included_fields:
            if isinstance(field, Query):
                nested_query_params = self.build_query_params(
                    field, parent=prefix + field.field_name
                )
                query_params.update(nested_query_params)
        return query_params

    def inject_query_params_in_req(self, request):
        parsed = self.get_parsed_restql_query(request)

        # Generate query params from query arguments
        query_params = self.build_query_params(parsed)

        # We are using `request.GET` instead of `request.query_params`
        # because at this point DRF request is not yet created so
        # `request.query_params` is not yet available
        params = request.GET.copy()
        params.update(query_params)

        # Make QueryDict immutable after updating
        request.GET = QueryDict(params.urlencode(), mutable=False)

    def dispatch(self, request, *args, **kwargs):
        self.inject_query_params_in_req(request)
        return super().dispatch(request, *args, **kwargs)


class DynamicFieldsMixin(RequestQueryParserMixin):
    def __init__(self, *args, **kwargs):
        # Don't pass DynamicFieldsMixin's kwargs to the superclass
        self.dynamic_fields_mixin_kwargs = {
            "query": kwargs.pop("query", None),
            "parsed_query": kwargs.pop("parsed_query", None),
            "fields": kwargs.pop("fields", None),
            "exclude": kwargs.pop("exclude", None),
            "return_pk": kwargs.pop("return_pk", False),
            "disable_dynamic_fields": kwargs.pop("disable_dynamic_fields", False),
        }

        msg = "May not set both `fields` and `exclude` kwargs"
        assert not (
            self.dynamic_fields_mixin_kwargs["fields"] is not None
            and self.dynamic_fields_mixin_kwargs["exclude"] is not None
        ), msg

        msg = "May not set both `query` and `parsed_query` kwargs"
        assert not (
            self.dynamic_fields_mixin_kwargs["query"] is not None
            and self.dynamic_fields_mixin_kwargs["parsed_query"] is not None
        ), msg

        # flag to toggle using restql fields
        self.is_ready_to_use_dynamic_fields = False

        # Instantiate the superclass normally
        super().__init__(*args, **kwargs)

    def to_representation(self, instance):
        # Activate using restql fields
        self.is_ready_to_use_dynamic_fields = True

        if self.dynamic_fields_mixin_kwargs["return_pk"]:
            return instance.pk
        return super().to_representation(instance)

    @cached_property
    def allowed_fields(self):
        fields = super().fields
        if self.dynamic_fields_mixin_kwargs["fields"] is not None:
            # Drop all fields which are not specified on the `fields` kwarg.
            allowed = set(self.dynamic_fields_mixin_kwargs["fields"])
            existing = set(fields)
            not_allowed = existing.symmetric_difference(allowed)
            for field_name in not_allowed:
                try:
                    fields.pop(field_name)
                except KeyError:
                    msg = "Field `%s` is not found" % field_name
                    raise FieldNotFound(msg) from None

        if self.dynamic_fields_mixin_kwargs["exclude"] is not None:
            # Drop all fields specified on the `exclude` kwarg.
            not_allowed = set(self.dynamic_fields_mixin_kwargs["exclude"])
            for field_name in not_allowed:
                try:
                    fields.pop(field_name)
                except KeyError:
                    msg = "Field `%s` is not found" % field_name
                    raise FieldNotFound(msg) from None
        return fields

    @staticmethod
    def is_field_found(field_name, all_field_names, raise_exception=False):
        if field_name in all_field_names:
            return True
        else:
            if raise_exception:
                msg = "`%s` field is not found" % field_name
                raise ValidationError(msg, code="not_found")
            return False

    @staticmethod
    def is_nested_field(field_name, field, raise_exception=False):
        nested_classes = (Serializer, ListSerializer, DynamicSerializerMethodField)
        if isinstance(field, nested_classes):
            return True
        else:
            if raise_exception:
                msg = "`%s` is not a nested field" % field_name
                raise ValidationError(msg, code="invalid")
            return False

    @staticmethod
    def is_valid_alias(alias):
        if len(alias) > restql_settings.MAX_ALIAS_LEN:
            msg = (
                "The length of `%s` alias has exceeded "
                "the limit specified, which is %s characters."
            ) % (alias, restql_settings.MAX_ALIAS_LEN)
            raise ValidationError(msg, code="invalid")

    def rename_aliased_fields(self, aliases, all_fields):
        for field, alias in aliases.items():
            self.is_field_found(field, all_fields, raise_exception=True)
            self.is_valid_alias(alias)
            all_fields[alias] = all_fields[field]
        return all_fields

    def select_fields(self, parsed_query, all_fields):
        self.rename_aliased_fields(parsed_query.aliases, all_fields)

        # The format is [field1, field2 ...]
        allowed_flat_fields = []

        # The format is  {nested_field: [sub_fields ...] ...}
        allowed_nested_fields = {}

        # The parsed_query.excluded_fields
        # is a list of names of excluded fields
        # The format is [field1, field2 ...]
        excluded_fields = parsed_query.excluded_fields

        # The parsed_query.included_fields
        # contains a list of allowed fields,
        # The format is [field, {nested_field: [sub_fields ...]} ...]
        included_fields = parsed_query.included_fields

        include_all_fields = False  # Assume the * is not set initially

        # Go through all included fields to check if
        # they are all valid and to set `nested_fields`
        # property on parent fields for future reference
        for field in included_fields:
            if field == "*":
                # Include all fields but ignore `*` since
                # it's not an actual field(it's just a flag)
                include_all_fields = True
                continue
            if isinstance(field, Query):
                # Nested field
                alias = parsed_query.aliases.get(field.field_name, field.field_name)

                self.is_field_found(field.field_name, all_fields, raise_exception=True)
                self.is_nested_field(
                    field.field_name, all_fields[field.field_name], raise_exception=True
                )
                allowed_nested_fields.update({alias: field})
            else:
                # Flat field
                alias = parsed_query.aliases.get(field, field)
                self.is_field_found(field, all_fields, raise_exception=True)
                allowed_flat_fields.append(alias)

        def get_duplicates(items):
            unique = []
            repeated = []
            for item in items:
                if item not in unique:
                    unique.append(item)
                else:
                    repeated.append(item)
            return repeated

        included_and_excluded_fields = (
            allowed_flat_fields + list(allowed_nested_fields.keys()) + excluded_fields
        )

        including_or_excluding_field_more_than_once = len(
            included_and_excluded_fields
        ) != len(set(included_and_excluded_fields))

        if including_or_excluding_field_more_than_once:
            repeated_fields = get_duplicates(included_and_excluded_fields)
            msg = (
                "QueryFormatError: You have either "
                "included/excluded a field more than once, "  # e.g {id, id}
                "used the same alias more than once, "  # e.g {x: name, x: age}
                "used a field name as an alias to another field or "  # e.g {id, id: age} Here age's not a parent
                "renamed a field and included/excluded it again, "  # e.g {ID: id, id}
                "The list of fields which led to this error is %s."
            ) % str(repeated_fields)
            raise ValidationError(msg, "invalid")

        if excluded_fields:
            # Here we are sure that parsed_query.excluded_fields
            # is not empty which means the user specified fields to exclude,
            # so we just check if provided fields exists then remove them from
            # a list of all fields
            for field in excluded_fields:
                self.is_field_found(field, all_fields, raise_exception=True)
                all_fields.pop(field)

        elif included_fields and not include_all_fields:
            # Here we are sure that parsed_query.excluded_fields
            # is empty which means the exclude operator(-) has not been used,
            # so parsed_query.included_fields contains only selected fields
            all_allowed_fields = set(allowed_flat_fields) | set(
                allowed_nested_fields.keys()
            )

            existing_fields = set(all_fields.keys())

            non_selected_fields = existing_fields - all_allowed_fields

            for field in non_selected_fields:
                # Remove it because we're sure it has not been selected
                all_fields.pop(field)

        elif include_all_fields:
            # Here we are sure both parsed_query.excluded_fields and
            # parsed_query.included_fields are empty, but * has been
            # used to select all fields, so we return all fields without
            # removing any
            pass

        else:
            # Otherwise the user specified empty query i.e query={}
            # So we return nothing
            all_fields = {}

        return all_fields, allowed_nested_fields

    @cached_property
    def dynamic_fields(self):
        parsed_restql_query = None

        is_root_serializer = self.parent is None or (
            isinstance(self.parent, ListSerializer) and self.parent.parent is None
        )

        if is_root_serializer:
            try:
                parsed_restql_query = self.get_parsed_restql_query()
            except SyntaxError as e:
                msg = "QuerySyntaxError: " + e.msg + " on " + e.text
                raise ValidationError(msg, code="invalid") from None
            except QueryFormatError as e:
                msg = "QueryFormatError: " + str(e)
                raise ValidationError(msg, code="invalid") from None

        elif isinstance(self.parent, ListSerializer):
            field_name = self.parent.field_name
            parent = self.parent.parent
            if hasattr(parent, "restql_nested_parsed_queries"):
                parent_nested_fields = parent.restql_nested_parsed_queries
                parsed_restql_query = parent_nested_fields.get(field_name, None)
        elif isinstance(self.parent, Serializer):
            field_name = self.field_name
            parent = self.parent
            if hasattr(parent, "restql_nested_parsed_queries"):
                parent_nested_fields = parent.restql_nested_parsed_queries
                parsed_restql_query = parent_nested_fields.get(field_name, None)

        if parsed_restql_query is None:
            # There's no query so we return all fields
            return self.allowed_fields

        # Get fields selected by `query` parameter
        selected_fields, nested_parsed_queries = self.select_fields(
            parsed_query=parsed_restql_query, all_fields=self.allowed_fields
        )

        # Keep track of parsed queries of nested fields
        # for future reference from child/nested serializers
        self.restql_nested_parsed_queries = nested_parsed_queries
        return selected_fields

    def get_parsed_restql_query_from_query_kwarg(self):
        parser = QueryParser()
        return parser.parse(self.dynamic_fields_mixin_kwargs["query"])

    def get_parsed_restql_query(self):
        request = self.context.get("request")

        if self.dynamic_fields_mixin_kwargs["query"] is not None:
            # Get from query kwarg
            return self.get_parsed_restql_query_from_query_kwarg()
        elif self.dynamic_fields_mixin_kwargs["parsed_query"] is not None:
            # Get from parsed_query kwarg
            return self.dynamic_fields_mixin_kwargs["parsed_query"]
        elif request is not None and self.has_restql_query_param(request):
            # Get from request query parameter
            return self.get_parsed_restql_query_from_req(request)
        return None  # There is no query so we return None as a parsed query

    @property
    def fields(self):
        should_use_dynamic_fields = (
            self.is_ready_to_use_dynamic_fields
            and not self.dynamic_fields_mixin_kwargs["disable_dynamic_fields"]
        )

        if should_use_dynamic_fields:
            # Return restql fields
            return self.dynamic_fields
        return self.allowed_fields


class EagerLoadingMixin(RequestQueryParserMixin):
    @property
    def parsed_restql_query(self):
        """
        Gets parsed query for use in eager loading.
        Defaults to the serializer parsed query.
        """
        if self.has_restql_query_param(self.request):
            try:
                return self.get_parsed_restql_query_from_req(self.request)
            except (SyntaxError, QueryFormatError):
                # Let `DynamicFieldsMixin` handle this for a user
                # to get a helpful error message
                pass

        # Else include all fields
        query = Query(
            field_name=None,
            included_fields=["*"],
            excluded_fields=[],
            aliases={},
            arguments={},
        )
        return query

    @property
    def should_auto_apply_eager_loading(self):
        if hasattr(self, "auto_apply_eager_loading"):
            return self.auto_apply_eager_loading
        return restql_settings.AUTO_APPLY_EAGER_LOADING

    def get_select_related_mapping(self):
        if hasattr(self, "select_related"):
            return self.select_related
        # Else select nothing
        return {}

    def get_prefetch_related_mapping(self):
        if hasattr(self, "prefetch_related"):
            return self.prefetch_related
        # Else prefetch nothing
        return {}

    @classmethod
    def get_dict_parsed_restql_query(cls, parsed_restql_query):
        """
        Returns the parsed query as a dict.
        """
        parsed_query = {}
        included_fields = parsed_restql_query.included_fields
        excluded_fields = parsed_restql_query.excluded_fields

        for field in included_fields:
            if isinstance(field, Query):
                nested_keys = cls.get_dict_parsed_restql_query(field)
                parsed_query[field.field_name] = nested_keys
            else:
                parsed_query[field] = True
        for field in excluded_fields:
            if isinstance(field, Query):
                nested_keys = cls.get_dict_parsed_restql_query(field)
                parsed_query[field.field_name] = nested_keys
            else:
                parsed_query[field] = False
        return parsed_query

    @staticmethod
    def get_related_fields(related_fields_mapping, dict_parsed_restql_query):
        """
        Returns only whitelisted related fields from a query to be used on
        `select_related` and `prefetch_related`
        """
        related_fields = []
        for key, related_field in related_fields_mapping.items():
            fields = key.split(".")
            if isinstance(related_field, (str, Prefetch)):
                related_field = [related_field]

            query_node = dict_parsed_restql_query
            for field in fields:
                if isinstance(query_node, dict):
                    if field in query_node:
                        # Get a more specific query node
                        query_node = query_node[field]
                    elif "*" in query_node:
                        # All fields are included
                        continue
                    else:
                        # The field is not included in a query so
                        # don't include this field in `related_fields`
                        break
            else:
                # If the loop completed without breaking
                if isinstance(query_node, dict) or query_node:
                    related_fields.extend(related_field)
        return related_fields

    def apply_eager_loading(self, queryset):
        """
        Applies appropriate select_related and prefetch_related calls on a
        queryset
        """
        query = self.get_dict_parsed_restql_query(self.parsed_restql_query)
        select_mapping = self.get_select_related_mapping()
        prefetch_mapping = self.get_prefetch_related_mapping()

        to_select = self.get_related_fields(select_mapping, query)
        to_prefetch = self.get_related_fields(prefetch_mapping, query)

        if to_select:
            queryset = queryset.select_related(*to_select)
        if to_prefetch:
            queryset = queryset.prefetch_related(*to_prefetch)
        return queryset

    def get_eager_queryset(self, queryset):
        return self.apply_eager_loading(queryset)

    def get_queryset(self):
        """
        Override for DRF's get_queryset on the view.
        If get_queryset is not present, we don't try to run this.
        Instead, this can still be used by manually calling
        self.get_eager_queryset and passing in the queryset desired.
        """
        if hasattr(super(), "get_queryset"):
            queryset = super().get_queryset()
            if self.should_auto_apply_eager_loading:
                queryset = self.get_eager_queryset(queryset)
            return queryset


class BaseNestedMixin(object):
    def get_fields(self):
        # Replace all temporary fields with the actual fields
        fields = super().get_fields()
        for field_name, field in fields.items():
            if isinstance(field, TemporaryNestedField):
                fields.update(
                    {field_name: field.get_actual_nested_field(self.__class__)}
                )
        return fields

    @cached_property
    def restql_writable_nested_fields(self):
        # Make field_source -> field_value map for restql nested fields
        writable_nested_fields = {}
        for _, field in self.fields.items():
            # Get the actual source of the field
            if isinstance(field, BaseRESTQLNestedField):
                writable_nested_fields.update({field.source: field})
        return writable_nested_fields


class NestedCreateMixin(BaseNestedMixin):
    """Create Mixin"""

    def create_writable_foreignkey_related(self, data):
        # data format
        # {field: {sub_field: value}}
        objs = {}
        nested_fields = self.restql_writable_nested_fields
        for field, value in data.items():
            # Get nested field serializer
            nested_field_serializer = nested_fields[field]
            serializer_class = nested_field_serializer.serializer_class
            kwargs = nested_field_serializer.validation_kwargs
            serializer = serializer_class(
                **kwargs,
                data=value,
                # Reject partial update by default(if partial kwarg is not passed)
                # since we need all required fields when creating object
                partial=nested_field_serializer.is_partial(False),
                context={**self.context, "parent_operation": CREATE},
            )
            serializer.is_valid(raise_exception=True)
            if value is None:
                objs.update({field: None})
            else:
                obj = serializer.save()
                objs.update({field: obj})
        return objs

    def bulk_create_objs(self, field, data):
        nested_fields = self.restql_writable_nested_fields

        # Get nested field serializer
        nested_field_serializer = nested_fields[field].child
        serializer_class = nested_field_serializer.serializer_class
        kwargs = nested_field_serializer.validation_kwargs
        pks = []
        for values in data:
            serializer = serializer_class(
                **kwargs,
                data=values,
                # Reject partial update by default(if partial kwarg is not passed)
                # since we need all required fields when creating object
                partial=nested_field_serializer.is_partial(False),
                context={**self.context, "parent_operation": CREATE},
            )
            serializer.is_valid(raise_exception=True)
            obj = serializer.save()
            pks.append(obj.pk)
        return pks

    def create_many_to_one_related(self, instance, data):
        # data format
        # {field: {
        # ADD: [pks],
        # CREATE: [{sub_field: value}]
        # }...}
        field_pks = {}
        for field, values in data.items():
            model = self.Meta.model
            foreignkey = getattr(model, field).field.name
            nested_fields = self.restql_writable_nested_fields
            for operation in values:
                if operation == ADD:
                    pks = values[operation]
                    model = nested_fields[field].child.Meta.model
                    qs = model.objects.filter(pk__in=pks)
                    qs.update(**{foreignkey: instance.pk})
                    field_pks.update({field: pks})
                elif operation == CREATE:
                    for v in values[operation]:
                        v.update({foreignkey: instance.pk})
                    pks = self.bulk_create_objs(field, values[operation])
                    field_pks.update({field: pks})
        return field_pks

    def create_many_to_one_generic_related(self, instance, data):
        field_pks = {}
        nested_fields = self.restql_writable_nested_fields

        content_type = (
            ContentType.objects.get_for_model(instance) if ContentType else None
        )
        for field, values in data.items():
            relation = getattr(self.Meta.model, field).field

            nested_field_serializer = nested_fields[field].child
            serializer_class = nested_field_serializer.serializer_class
            kwargs = nested_field_serializer.validation_kwargs
            model = nested_field_serializer.Meta.model

            for operation in values:
                if operation == ADD:
                    pks = values[operation]
                    qs = model.objects.filter(pk__in=pks)
                    qs.update(
                        **{
                            relation.object_id_field_name: instance.pk,
                            relation.content_type_field_name: content_type,
                        }
                    )
                elif operation == CREATE:
                    serializer = serializer_class(
                        data=values[operation], **kwargs, many=True
                    )
                    serializer.is_valid(raise_exception=True)
                    items = serializer.validated_data

                    objs = [
                        model(
                            **item,
                            **{
                                relation.content_type_field_name: content_type,
                                relation.object_id_field_name: instance.pk,
                            },
                        )
                        for item in items
                    ]
                    objs = model.objects.bulk_create(objs)
                    field_pks[field] = [obj.pk for obj in objs]

        return field_pks

    def create_many_to_many_related(self, instance, data):
        # data format
        # {field: {
        # ADD: [pks],
        # CREATE: [{sub_field: value}]
        # }...}
        field_pks = {}
        for field, values in data.items():
            obj = getattr(instance, field)
            for operation in values:
                if operation == ADD:
                    pks = values[operation]
                    obj.add(*pks)
                    field_pks.update({field: pks})
                elif operation == CREATE:
                    pks = self.bulk_create_objs(field, values[operation])
                    obj.add(*pks)
                    field_pks.update({field: pks})
        return field_pks

    def create(self, validated_data):
        # Make a copy of validated_data so that we don't
        # alter it in case user need to access it later
        validated_data_copy = {**validated_data}

        fields = {
            "foreignkey_related": {"replaceable": {}, "writable": {}},
            "many_to": {
                "many_related": {},
                "one_related": {},
                "one_generic_related": {},
            },
        }

        restql_nested_fields = self.restql_writable_nested_fields
        for field in restql_nested_fields:
            if field not in validated_data_copy:
                # Nested field value is not provided
                continue

            field_serializer = restql_nested_fields[field]

            if isinstance(field_serializer, Serializer):
                if field_serializer.is_replaceable:
                    value = validated_data_copy.pop(field)
                    fields["foreignkey_related"]["replaceable"].update({field: value})
                else:
                    value = validated_data_copy.pop(field)
                    fields["foreignkey_related"]["writable"].update({field: value})
            elif isinstance(field_serializer, ListSerializer):
                model = self.Meta.model
                rel = getattr(model, field).rel

                if isinstance(rel, ManyToOneRel):
                    value = validated_data_copy.pop(field)
                    fields["many_to"]["one_related"].update({field: value})
                elif isinstance(rel, ManyToManyRel):
                    value = validated_data_copy.pop(field)
                    fields["many_to"]["many_related"].update({field: value})
                elif GenericRel and isinstance(rel, GenericRel):
                    value = validated_data_copy.pop(field)
                    fields["many_to"]["one_generic_related"].update({field: value})

        foreignkey_related = {
            **fields["foreignkey_related"]["replaceable"],
            **self.create_writable_foreignkey_related(
                fields["foreignkey_related"]["writable"]
            ),
        }

        instance = super().create({**validated_data_copy, **foreignkey_related})

        self.create_many_to_many_related(instance, fields["many_to"]["many_related"])

        self.create_many_to_one_related(instance, fields["many_to"]["one_related"])

        if fields["many_to"]["one_generic_related"]:
            # Call create_many_to_one_generic_related only if we have generic relationship
            self.create_many_to_one_generic_related(
                instance, fields["many_to"]["one_generic_related"]
            )

        return instance


class NestedUpdateMixin(BaseNestedMixin):
    """Update Mixin"""

    @staticmethod
    def constrain_error_prefix(field):
        return "Error on `%s` field: " % (field,)

    @staticmethod
    def update_replaceable_foreignkey_related(instance, data):
        # data format {field: obj}
        for field, nested_obj in data.items():
            setattr(instance, field, nested_obj)
        if data:
            instance.save()

    def update_writable_foreignkey_related(self, instance, data):
        # data format {field: {sub_field: value}}
        nested_fields = self.restql_writable_nested_fields

        needs_save = False
        for field, values in data.items():
            # Get nested field serializer
            nested_field_serializer = nested_fields[field]
            serializer_class = nested_field_serializer.serializer_class
            kwargs = nested_field_serializer.validation_kwargs
            nested_obj = getattr(instance, field)
            serializer = serializer_class(
                nested_obj,
                **kwargs,
                data=values,
                # Allow partial update by default(if partial kwarg is not passed)
                # since this is nested update
                partial=nested_field_serializer.is_partial(True),
                context={**self.context, "parent_operation": UPDATE},
            )
            serializer.is_valid(raise_exception=True)
            if values is None:
                setattr(instance, field, None)
                needs_save = True
            else:
                obj = serializer.save()
                if nested_obj is None:
                    # Patch back newly created object to instance
                    setattr(instance, field, obj)
                    needs_save = True
        if needs_save:
            instance.save()

    def bulk_create_many_to_many_related(self, field, nested_obj, data):
        # Get nested field serializer
        nested_field_serializer = self.restql_writable_nested_fields[field].child
        serializer_class = nested_field_serializer.serializer_class
        kwargs = nested_field_serializer.validation_kwargs
        pks = []
        for values in data:
            serializer = serializer_class(
                **kwargs,
                data=values,
                # Reject partial update by default(if partial kwarg is not passed)
                # since we need all required fields when creating object
                partial=nested_field_serializer.is_partial(False),
                context={**self.context, "parent_operation": CREATE},
            )
            serializer.is_valid(raise_exception=True)
            obj = serializer.save()
            pks.append(obj.pk)
        nested_obj.add(*pks)
        return pks

    def bulk_create_many_to_one_related(self, field, nested_obj, data):
        # Get nested field serializer
        nested_field_serializer = self.restql_writable_nested_fields[field].child
        serializer_class = nested_field_serializer.serializer_class
        kwargs = nested_field_serializer.validation_kwargs
        pks = []
        for values in data:
            serializer = serializer_class(
                **kwargs,
                data=values,
                # Reject partial update by default(if partial kwarg is not passed)
                # since we need all required fields when creating object
                partial=nested_field_serializer.is_partial(False),
                context={**self.context, "parent_operation": CREATE},
            )
            serializer.is_valid(raise_exception=True)
            obj = serializer.save()
            pks.append(obj.pk)
        return pks

    def bulk_update_many_to_many_related(self, field, nested_obj, data):
        # {pk: {sub_field: values}}

        # Get nested field serializer
        nested_field_serializer = self.restql_writable_nested_fields[field].child
        serializer_class = nested_field_serializer.serializer_class
        kwargs = nested_field_serializer.validation_kwargs
        for pk, values in data.items():
            try:
                obj = nested_obj.get(pk=pk)
            except ObjectDoesNotExist:
                # This pk does't belong to nested field
                continue
            serializer = serializer_class(
                obj,
                **kwargs,
                data=values,
                # Allow partial update by default(if partial kwarg is not passed)
                # since this is nested update
                partial=nested_field_serializer.is_partial(True),
                context={**self.context, "parent_operation": UPDATE},
            )
            serializer.is_valid(raise_exception=True)
            obj = serializer.save()

    def bulk_update_many_to_one_related(
        self, field, instance, data, update_foreign_key=True
    ):
        # {pk: {sub_field: values}}

        # Get nested field serializer
        nested_field_serializer = self.restql_writable_nested_fields[field].child
        serializer_class = nested_field_serializer.serializer_class
        kwargs = nested_field_serializer.validation_kwargs
        model = self.Meta.model
        foreignkey = getattr(model, field).field.name
        nested_obj = getattr(instance, field)
        for pk, values in data.items():
            try:
                obj = nested_obj.get(pk=pk)
            except ObjectDoesNotExist:
                # This pk does't belong to nested field
                continue
            if update_foreign_key:
                values.update({foreignkey: instance.pk})
            serializer = serializer_class(
                obj,
                **kwargs,
                data=values,
                # Allow partial update by default(if partial kwarg is not passed)
                # since this is nested update
                partial=nested_field_serializer.is_partial(True),
                context={**self.context, "parent_operation": UPDATE},
            )
            serializer.is_valid(raise_exception=True)
            obj = serializer.save()

    def update_many_to_one_related(self, instance, data):
        # data format
        # {field: {
        # ADD: [{sub_field: value}],
        # CREATE: [{sub_field: value}],
        # REMOVE: [pk],
        # UPDATE: {pk: {sub_field: value}}
        # }...}
        for field, values in data.items():
            nested_obj = getattr(instance, field)
            model = self.Meta.model
            foreignkey = getattr(model, field).field.name
            nested_fields = self.restql_writable_nested_fields
            for operation in values:
                if operation == ADD:
                    pks = values[operation]
                    model = nested_fields[field].child.Meta.model
                    qs = model.objects.filter(pk__in=pks)
                    qs.update(**{foreignkey: instance.pk})
                elif operation == CREATE:
                    for v in values[operation]:
                        v.update({foreignkey: instance.pk})
                    self.bulk_create_many_to_one_related(
                        field, nested_obj, values[operation]
                    )
                elif operation == REMOVE:
                    qs = nested_obj.all()
                    if values[operation] == ALL_RELATED_OBJS:
                        qs.delete()
                    else:
                        qs.filter(pk__in=values[operation]).delete()
                elif operation == UPDATE:
                    self.bulk_update_many_to_one_related(
                        field, instance, values[operation]
                    )
                else:
                    message = "`%s` is an invalid operation" % (operation,)
                    raise ValidationError(message, code="invalid_operation")
        return instance

    def update_many_to_one_generic_related(self, instance, data):
        # it's same logic for add & create operations
        # which are already handled by NestedCreateMixin
        NestedCreateMixin.create_many_to_one_generic_related(self, instance, data)

        for field, values in data.items():
            nested_qs = getattr(instance, field)
            for operation in values:
                if operation not in [ADD, CREATE, UPDATE, REMOVE]:
                    message = f"`{operation}` is an invalid operation"
                    raise ValidationError(message, code="invalid_operation")

                if operation == REMOVE:
                    qs = nested_qs.all()
                    if values[operation] == ALL_RELATED_OBJS:
                        qs.delete()
                    else:
                        qs.filter(pk__in=values[operation]).delete()
                elif operation == UPDATE:
                    self.bulk_update_many_to_one_related(
                        field, instance, values[operation], update_foreign_key=False
                    )
        return instance

    def update_many_to_many_related(self, instance, data):
        # data format
        # {field: {
        # ADD: [{sub_field: value}],
        # CREATE: [{sub_field: value}],
        # REMOVE: [pk],
        # UPDATE: {pk: {sub_field: value}}
        # }...}
        for field, values in data.items():
            nested_obj = getattr(instance, field)
            for operation in values:
                if operation == ADD:
                    pks = values[operation]
                    try:
                        nested_obj.add(*pks)
                    except Exception as e:
                        msg = self.constrain_error_prefix(field) + str(e)
                        code = "constrain_error"
                        raise ValidationError(msg, code=code) from None
                elif operation == CREATE:
                    self.bulk_create_many_to_many_related(
                        field, nested_obj, values[operation]
                    )
                elif operation == REMOVE:
                    pks = values[operation]
                    if pks == ALL_RELATED_OBJS:
                        pks = nested_obj.all()
                    try:
                        nested_obj.remove(*pks)
                    except Exception as e:
                        msg = self.constrain_error_prefix(field) + str(e)
                        code = "constrain_error"
                        raise ValidationError(msg, code=code) from None
                elif operation == UPDATE:
                    self.bulk_update_many_to_many_related(
                        field, nested_obj, values[operation]
                    )
                else:
                    message = "`%s` is an invalid operation" % (operation,)
                    raise ValidationError(message, code="invalid_operation")
        return instance

    def update(self, instance, validated_data):
        # Make a copty of validated_data so that we don't
        # alter it in case user need to access it later
        validated_data_copy = {**validated_data}

        fields = {
            "foreignkey_related": {"replaceable": {}, "writable": {}},
            "many_to": {
                "many_related": {},
                "one_related": {},
                "one_generic_related": {},
            },
        }

        restql_nested_fields = self.restql_writable_nested_fields
        for field in restql_nested_fields:
            if field not in validated_data_copy:
                # Nested field value is not provided
                continue

            field_serializer = restql_nested_fields[field]

            if isinstance(field_serializer, Serializer):
                if field_serializer.is_replaceable:
                    value = validated_data_copy.pop(field)
                    fields["foreignkey_related"]["replaceable"].update({field: value})
                else:
                    value = validated_data_copy.pop(field)
                    fields["foreignkey_related"]["writable"].update({field: value})
            elif isinstance(field_serializer, ListSerializer):
                model = self.Meta.model
                rel = getattr(model, field).rel

                if isinstance(rel, ManyToOneRel):
                    value = validated_data_copy.pop(field)
                    fields["many_to"]["one_related"].update({field: value})
                elif isinstance(rel, ManyToManyRel):
                    value = validated_data_copy.pop(field)
                    fields["many_to"]["many_related"].update({field: value})
                elif GenericRel and isinstance(rel, GenericRel):
                    value = validated_data_copy.pop(field)
                    fields["many_to"]["one_generic_related"].update({field: value})

        instance = super().update(instance, validated_data_copy)

        self.update_replaceable_foreignkey_related(
            instance, fields["foreignkey_related"]["replaceable"]
        )

        self.update_writable_foreignkey_related(
            instance, fields["foreignkey_related"]["writable"]
        )

        self.update_many_to_many_related(instance, fields["many_to"]["many_related"])

        self.update_many_to_one_related(instance, fields["many_to"]["one_related"])

        if fields["many_to"]["one_generic_related"]:
            # Call update_many_to_one_generic_related only if we have generic relationship
            self.update_many_to_one_generic_related(
                instance, fields["many_to"]["one_generic_related"]
            )
        return instance



import re
from collections import namedtuple

from pypeg2 import List, contiguous, csl, name, optional, parse

from .exceptions import QueryFormatError


class Alias(List):
    grammar = name(), ':'


class IncludedField(List):
    grammar = optional(Alias), name()

    @property
    def alias(self):
        if len(self) > 0:
            return self[0].name
        return None


class ExcludedField(List):
    grammar = contiguous('-', name())


class AllFields(str):
    grammar = '*'


class ArgumentWithoutQuotes(List):
    grammar = name(), ':', re.compile(r'true|false|null|[-+]?[0-9]*\.?[0-9]+')

    def number(self, val):
        try:
            return int(val)
        except ValueError:
            return float(val)

    @property
    def value(self):
        raw_val = self[0]
        FIXED_DATA_TYPES = {
            'true': True,
            'false': False,
            'null': None
        }
        if raw_val in FIXED_DATA_TYPES:
            return FIXED_DATA_TYPES[raw_val]
        return self.number(raw_val)


class ArgumentWithQuotes(List):
    grammar = name(), ':', re.compile(r'"([^"\\]|\\.|\\\n)*"|\'([^\'\\]|\\.|\\\n)*\'')

    @property
    def value(self):
        # Slicing is for removing quotes
        # at the begining and end of a string
        return self[0][1:-1]


class Arguments(List):
    grammar = optional(csl(
        [
            ArgumentWithoutQuotes,
            ArgumentWithQuotes,
        ],
        separator=[',', '']
    ))


class ArgumentsBlock(List):
    grammar = optional('(', Arguments, optional(','), ')')

    @property
    def arguments(self):
        if self[0] is None:
            return []  # No arguments
        return self[0]


class ParentField(List):
    """
    According to ParentField grammar:
    self[0]  returns IncludedField instance,
    self[1]  returns Block instance
    """
    @property
    def name(self):
        return self[0].name

    @property
    def alias(self):
        return self[0].alias

    @property
    def block(self):
        return self[1]


class BlockBody(List):
    grammar = optional(csl(
        [ParentField, IncludedField, ExcludedField, AllFields],
        separator=[',', '']
    ))


class Block(List):
    grammar = ArgumentsBlock, '{', BlockBody, optional(','), '}'

    @property
    def arguments(self):
        return self[0].arguments

    @property
    def body(self):
        return self[1]


# ParentField grammar,
# We don't include `ExcludedField` here because
# exclude operator(-) on a parent field should
# raise syntax error, e.g {name, -location{city}}
# `IncludedField` is a parent field and `Block`
# contains its sub fields
ParentField.grammar = IncludedField, Block


Query = namedtuple(
    "Query",
    ("field_name", "included_fields", "excluded_fields", "aliases", "arguments")
)


class QueryParser(object):
    def parse(self, query):
        parse_tree = parse(query, Block)
        return self._transform_block(parse_tree, parent_field=None)

    def _transform_block(self, block, parent_field=None):
        query = Query(
            field_name=parent_field,
            included_fields=[],
            excluded_fields=[],
            aliases={},
            arguments={}
        )

        for argument in block.arguments:
            argument = {str(argument.name): argument.value}
            query.arguments.update(argument)

        for field in block.body:
            # A field may be a parent or included field or excluded field
            if isinstance(field, (ParentField, IncludedField)):
                # Find all aliases
                if field.alias:
                    query.aliases.update({str(field.name): str(field.alias)})

            field = self._transform_field(field)

            if isinstance(field, Query):
                # A field is a parent
                query.included_fields.append(field)
            elif isinstance(field, IncludedField):
                query.included_fields.append(str(field.name))
            elif isinstance(field, ExcludedField):
                query.excluded_fields.append(str(field.name))
            elif isinstance(field, AllFields):
                # include all fields
                query.included_fields.append("*")

        if query.excluded_fields and "*" not in query.included_fields:
            query.included_fields.append("*")

        field_names = set(query.aliases.values())
        field_aliases = set(query.aliases.keys())
        faulty_fields = field_names.intersection(field_aliases)
        if faulty_fields:
            # We check this here because if we let it pass during
            # parsing it's going to raise inappropriate error message
            # when checking fields availability(for the case of renamed parents)
            msg = (
                "You have either "
                "used an existing field name as an alias to another field or "  # e.g {id, id: course{}}
                "you have defined an alias with the same name as a field name."  # e.g {id: id}
                "The list of fields which led to this error is %s."
            ) % str(list(faulty_fields))
            raise QueryFormatError(msg)
        return query

    def _transform_field(self, field):
        # A field may be a parent or included field or excluded field
        if isinstance(field, ParentField):
            return self._transform_parent_field(field)
        return field

    def _transform_parent_field(self, parent_field):
        return self._transform_block(
            parent_field.block,
            parent_field=str(parent_field.name)
        )



try:
    from django.utils.decorators import classproperty
except ImportError:
    from django.utils.functional import classproperty
from django.db.models.fields.related import ManyToOneRel

from rest_framework.fields import (
    DictField, ListField, SkipField, Field, empty
)
from rest_framework.serializers import (
    ListSerializer, PrimaryKeyRelatedField,
    SerializerMethodField, ValidationError
)

from .parser import Query
from .exceptions import InvalidOperation
from .operations import ADD, CREATE, REMOVE, UPDATE

CREATE_OPERATIONS = (ADD, CREATE)
UPDATE_OPERATIONS = (ADD, CREATE, REMOVE, UPDATE)

ALL_RELATED_OBJS = '__all__'


class DynamicSerializerMethodField(SerializerMethodField):
    def to_representation(self, value):
        method = getattr(self.parent, self.method_name)
        is_parsed_query_available = (
            hasattr(self.parent, "restql_nested_parsed_queries") and
            self.field_name in self.parent.restql_nested_parsed_queries
        )

        if is_parsed_query_available:
            parsed_query = self.parent.restql_nested_parsed_queries[self.field_name]
        else:
            # Include all fields
            parsed_query = Query(
                field_name=None,
                included_fields=["*"],
                excluded_fields=[],
                aliases={},
                arguments={}
            )
        return method(value, parsed_query)


class BaseRESTQLNestedField(object):
    def to_internal_value(self, data):
        raise NotImplementedError('`to_internal_value()` must be implemented.')


def BaseNestedFieldSerializerFactory(
        *args,
        accept_pk=False,
        accept_pk_only=False,
        allow_remove_all=False,
        create_ops=CREATE_OPERATIONS,
        update_ops=UPDATE_OPERATIONS,
        serializer_class=None,
        **kwargs):
    many = kwargs.get("many", False)
    partial = kwargs.get("partial", None)

    assert not (
        many and (accept_pk or accept_pk_only)
    ), (
        "May not set both `many=True` and `accept_pk=True` "
        "or `accept_pk_only=True`"
        "(accept_pk and accept_pk_only applies to foreign key relation only)."
    )

    assert not (
        accept_pk and accept_pk_only
    ), "May not set both `accept_pk=True` and `accept_pk_only=True`"

    assert not (
        allow_remove_all and not many
    ), (
        "`allow_remove_all=True` can only be applied to many related "
        "nested fields, ensure the kwarg `many=True` is set."
    )

    def join_words(words, many='are', single='is'):
        word_list = ["`" + word + "`" for word in words]

        if len(words) > 1:
            sentence = " & ".join([", ".join(word_list[:-1]), word_list[-1]])
            return "%s %s" % (many, sentence)
        elif len(words) == 1:
            return "%s %s" % (single, word_list[0])
        return "%s %s" % (single, "[]")

    if not set(create_ops).issubset(set(CREATE_OPERATIONS)):
        msg = (
            "Invalid create operation(s) at `%s`, Supported operations " +
            join_words(CREATE_OPERATIONS)
        ) % "create_ops=%s" % create_ops
        raise InvalidOperation(msg)

    if not set(update_ops).issubset(set(UPDATE_OPERATIONS)):
        msg = (
            "Invalid update operation(s) at `%s`, Supported operations " +
            join_words(UPDATE_OPERATIONS)
        ) % "update_ops=%s" % update_ops
        raise InvalidOperation(msg)

    if serializer_class == "self":
        # We have a self referencing serializer so the serializer
        # class is not available at the moment, we return None
        return None

    class BaseNestedField(BaseRESTQLNestedField):
        @classproperty
        def serializer_class(cls):
            # Return original nested serializer
            return serializer_class

        def is_partial(self, default):
            # Check if partial kwarg is passed if not return the default
            if partial is not None:
                return partial
            return default

    class BaseNestedFieldListSerializer(ListSerializer, BaseNestedField):
        def run_pk_list_validation(self, pks):
            ListField().run_validation(pks)
            queryset = self.child.Meta.model.objects.all()
            PrimaryKeyRelatedField(
                **self.child.validation_kwargs,
                queryset=queryset,
                many=True
            ).run_validation(pks)

        def run_data_list_validation(self, data, partial=None, operation=None):
            ListField().run_validation(data)
            model = self.parent.Meta.model
            rel = getattr(model, self.source).rel
            if isinstance(rel, ManyToOneRel):
                # ManyToOne Relation
                field_name = getattr(model, self.source).field.name
                child_serializer = serializer_class(
                    **self.child.validation_kwargs,
                    data=data,
                    many=True,
                    partial=partial,
                    context={**self.context, "parent_operation": operation}
                )

                # Remove parent field(field_name) for validation purpose
                child_serializer.child.fields.pop(field_name, None)

                # Check if a serializer is valid
                child_serializer.is_valid(raise_exception=True)
            else:
                # ManyToMany Relation
                child_serializer = serializer_class(
                    **self.child.validation_kwargs,
                    data=data,
                    many=True,
                    partial=partial,
                    context={**self.context, "parent_operation": operation}
                )

                # Check if a serializer is valid
                child_serializer.is_valid(raise_exception=True)

        def run_add_list_validation(self, data):
            self.run_pk_list_validation(data)

        def run_create_list_validation(self, data):
            self.run_data_list_validation(
                data,
                partial=self.is_partial(False),
                operation=CREATE
            )

        def run_remove_list_validation(self, data):
            if data == ALL_RELATED_OBJS:
                if not allow_remove_all:
                    msg = (
                        "Using `%s` value on `%s` operation is disabled"
                        % (ALL_RELATED_OBJS, REMOVE)
                    )
                    raise ValidationError(msg, code="not_allowed")
            else:
                self.run_pk_list_validation(data)

        def run_update_list_validation(self, data):
            DictField().run_validation(data)
            pks = list(data.keys())
            self.run_pk_list_validation(pks)
            values = list(data.values())
            self.run_data_list_validation(
                values,
                partial=self.is_partial(True),
                operation=UPDATE
            )

        def run_data_validation(self, data, allowed_ops):
            DictField().run_validation(data)

            operation_2_validation_method = {
                ADD: self.run_add_list_validation,
                CREATE: self.run_create_list_validation,
                REMOVE: self.run_remove_list_validation,
                UPDATE: self.run_update_list_validation,
            }

            allowed_operation_2_validation_method = {
                operation: operation_2_validation_method[operation]
                for operation in allowed_ops
            }

            for operation, values in data.items():
                try:
                    allowed_operation_2_validation_method[operation](values)
                except ValidationError as e:
                    detail = {operation: e.detail}
                    code = e.get_codes()
                    raise ValidationError(detail, code=code) from None
                except KeyError:
                    msg = (
                        "`%s` is not a valid operation, valid operations(s) "
                        "for this request %s"
                        % (operation, join_words(allowed_ops))
                    )
                    code = 'invalid_operation'
                    raise ValidationError(msg, code=code) from None

        def to_internal_value(self, data):
            if self.child.root.instance is None:
                parent_operation = self.context.get("parent_operation")
                if parent_operation == "update":
                    # Definitely an update
                    self.run_data_validation(data, update_ops)
                else:
                    self.run_data_validation(data, create_ops)
            else:
                # Definitely an update
                self.run_data_validation(data, update_ops)
            return data

        def __repr__(self):
            return (
                "BaseNestedField(%s, many=True)" %
                (serializer_class.__name__, )
            )

    class BaseNestedFieldSerializer(serializer_class, BaseNestedField):

        # might be used before `to_internal_value` method is called
        # so we're creating this property to make sure it's available
        # as long as the class is created
        is_replaceable = accept_pk_only or accept_pk

        class Meta(serializer_class.Meta):
            list_serializer_class = BaseNestedFieldListSerializer

        def run_validation(self, data):
            # Run `to_internal_value` only nothing more
            # This is needed only on DRF 3.8.x due to a bug on it
            # This function can be removed on other supported DRF versions
            # i.e v3.7 v3.9 v3.10 etc doesn't need this function
            return self.to_internal_value(data)

        def run_pk_validation(self, pk):
            queryset = self.Meta.model.objects.all()
            validator = PrimaryKeyRelatedField(
                **self.validation_kwargs,
                queryset=queryset,
                many=False
            )
            # If valid return object instead of pk
            return validator.run_validation(pk)

        def run_data_validation(self, data):
            parent_operation = self.context.get("parent_operation")

            child_serializer = serializer_class(
                **self.validation_kwargs,
                data=data,
                partial=self.is_partial(
                    # Use the partial value passed, if it's not passed
                    # Use the one from the top level parent
                    True if parent_operation == UPDATE else False
                ),
                context=self.context
            )

            # Set parent to a child serializer
            child_serializer.parent = self.parent

            # Check if a serializer is valid
            child_serializer.is_valid(raise_exception=True)

            # return data to be passed to a nested serializer,
            # don't be tempted to return child_serializer.validated_data
            # cuz it changes representation of some values for instance
            # pks gets converted into objects
            return data

        def to_internal_value(self, data):
            required = kwargs.get('required', True)
            default = kwargs.get('default', empty)

            if data == empty:
                # Implementation under this block is made
                # according to DRF behaviour to other normal fields
                # For more details see
                # https://www.django-rest-framework.org/api-guide/fields/#required
                # https://www.django-rest-framework.org/api-guide/fields/#default
                # https://www.django-rest-framework.org/api-guide/fields/#allow_null
                if self.root.partial or not required:
                    # Skip the field because the update is partial
                    # or the field is not required(optional)
                    raise SkipField()
                elif required:
                    if default == empty:
                        raise ValidationError(
                            "This field is required.",
                            code='required'
                        )
                    else:
                        # Use the default value
                        data = default

            if accept_pk_only:
                return self.run_pk_validation(data)
            elif accept_pk:
                if isinstance(data, dict):
                    self.is_replaceable = False
                    return self.run_data_validation(data)
                else:
                    return self.run_pk_validation(data)
            return self.run_data_validation(data)

        def __repr__(self):
            return (
                "BaseNestedField(%s, many=False)" %
                (serializer_class.__name__, )
            )

    return {
        "serializer_class": BaseNestedFieldSerializer,
        "list_serializer_class": BaseNestedFieldListSerializer,
        "args": args,
        "kwargs": kwargs
    }


class TemporaryNestedField(Field, BaseRESTQLNestedField):
    """
    This is meant to be used temporarily when 'self' is
    passed as the first arg to `NestedField`
    """

    def __init__(
            self, NestedField, *args,
            field_args=None, field_kwargs=None, **kwargs):
        self.field_args = field_args
        self.field_kwargs = field_kwargs
        self.NestedField = NestedField
        super().__init__(*args, **kwargs)

    def get_actual_nested_field(self, serializer_class):
        # Replace "self" with the actual parent serializer class
        self.field_kwargs.update({
            "serializer_class": serializer_class
        })

        # Reproduce the actual field
        return self.NestedField(
            *self.field_args,
            **self.field_kwargs
        )


def NestedFieldWraper(*args, **kwargs):
    serializer_class = kwargs["serializer_class"]
    factory = BaseNestedFieldSerializerFactory(*args, **kwargs)

    if factory is None:
        # We have a self referencing serializer so we return
        # a temporary field while we are waiting for the parent
        # to be ready(when it's ready the parent itself will replace
        # this field with the actual field)
        return TemporaryNestedField(
            NestedFieldWraper,
            field_args=args,
            field_kwargs=kwargs
        )

    serializer_validation_kwargs = {**factory['kwargs']}

    # Remove all non validation related kwargs and
    # DynamicFieldsMixin kwargs from `valdation_kwargs`
    non_validation_related_kwargs = [
        'many', 'data', 'instance', 'context', 'fields',
        'exclude', 'return_pk', 'disable_dynamic_fields',
        'query', 'parsed_query', 'partial'
    ]

    for kwarg in non_validation_related_kwargs:
        serializer_validation_kwargs.pop(kwarg, None)

    class NestedListSerializer(factory["list_serializer_class"]):
        def __repr__(self):
            return (
                "NestedField(%s, many=False)" %
                (serializer_class.__name__, )
            )

    class NestedSerializer(factory["serializer_class"]):
        # set validation related kwargs to be used on
        # `NestedCreateMixin` and `NestedUpdateMixin`
        validation_kwargs = serializer_validation_kwargs

        class Meta(factory["serializer_class"].Meta):
            list_serializer_class = NestedListSerializer

        def __repr__(self):
            return (
                "NestedField(%s, many=False)" %
                (serializer_class.__name__, )
            )

    return NestedSerializer(
        *factory["args"],
        **factory["kwargs"]
    )


def NestedField(serializer_class, *args, **kwargs):
    return NestedFieldWraper(
        *args,
        **kwargs,
        serializer_class=serializer_class
    )
